# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from collections import OrderedDict
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)
from .base import (
    Connection,
    Element,
)
from .series import Series
from .parallel import Parallel
from .resistor import Resistor
from .capacitor import Capacitor
from .inductor import Inductor
from .constant_phase_element import ConstantPhaseElement
from numpy import (
    array,
    inf,
    ndarray,
)
from schemdraw import Drawing
import schemdraw.elements as elm
from sympy import (
    Expr,
    latex,
)


ElementParameters = Dict[str, Dict[str, str]]


class Circuit:
    """
    A class that represents an equivalent circuit.

    Parameters
    ----------
    elements: Series
        The elements of the circuit wrapped in a Series connection.

    label: str = ""
        The label assigned to the circuit.
    """

    def __init__(self, elements: Series, label: str = ""):
        assert type(elements) is Series
        self._elements: Series = elements
        self._assign_identifiers()
        self._label: str = label or self.to_string()

    def __repr__(self) -> str:
        return f"Circuit ('{self.to_string()}', {hex(id(self))})"

    def __str__(self) -> str:
        return self.to_string()

    def get_label(self) -> str:
        """
        Get the label assigned to this circuit.

        Returns
        -------
        str
        """
        return self._label

    def set_label(self, label: str):
        """
        Set the label assigned to this circuit.

        Parameters
        ----------
        label: str
            The new label.
        """
        assert type(label) is str
        self._label = label.strip()

    def _assign_identifiers(self):
        return self._elements._assign_identifier(0)

    def to_stack(self) -> List[Tuple[str, Union[Element, Connection]]]:
        stack: List[Tuple[str, Union[Element, Connection]]] = []
        self._elements.to_stack(stack)
        return stack

    def to_string(self, decimals: int = -1) -> str:
        """
        Generate the circuit description code (CDC) that represents this circuit.

        Parameters
        ----------
        decimals: int = -1
            The number of decimals to include for the current element parameter values and limits.
            -1 means that the CDC is generated using the basic syntax, which omits element labels, parameter values, and parameter limits.

        Returns
        -------
        str
        """
        return self._elements.to_string(decimals=decimals)

    def impedance(self, f: float) -> complex:
        """
        Calculate the impedance of this circuit at a single frequency.

        Parameters
        ----------
        f: float
            The frequency in hertz.

        Returns
        -------
        complex
        """
        assert f > 0 and f < inf
        return self._elements.impedance(f)

    def impedances(self, f: Union[list, ndarray]) -> ndarray:
        """
        Calculate the impedance of this circuit at multiple frequencies.

        Parameters
        ----------
        f: Union[list, ndarray]

        Returns
        -------
        ndarray
        """
        assert type(f) is list or type(f) is ndarray, f
        assert min(f) > 0 and max(f) < inf, f
        return array(list(map(self.impedance, f)))

    def get_connections(self, flattened: bool = True) -> List[Connection]:
        """
        Get the connections in this circuit.

        Parameters
        ----------
        flattened: bool = True
            Whether or not the connections should be returned as a list of all connections or as a list connections that may also contain more connections.

        Returns
        -------
        List[Connection]
        """
        if flattened is True:
            return self._elements.get_connections(flattened=flattened)
        return [self._elements]

    def get_elements(self, flattened: bool = True) -> List[Union[Element, Connection]]:
        """
        Get the elements in this circuit.

        Parameters
        ----------
        flattened: bool = True
            Whether or not the elements should be returned as a list of only elements or as a list of connections containing elements.

        Returns
        -------
        List[Union[Element, Connection]]
        """
        if flattened is True:
            return self._elements.get_elements(flattened=flattened)
        return [self._elements]

    def get_parameters(self) -> "Dict[int, OrderedDict[str, float]]":
        """
        Get a mapping of each circuit element's integer identifier to an OrderedDict representing that element's parameters.

        Returns
        -------
        Dict[int, OrderedDict[str, float]]
        """
        return self._elements.get_parameters()

    def set_parameters(self, parameters: Dict[int, Dict[str, float]]):
        """
        Assign new parameters to the circuit elements.

        Parameters
        ----------
        parameters: Dict[int, Dict[str, float]]
            A mapping of circuit element integer identifiers to an OrderedDict mapping the parameter symbol to the new value.
        """
        self._elements.set_parameters(parameters)

    def get_element(self, ident: int) -> Optional[Element]:
        """
        Get the circuit element with a given integer identifier.

        Parameters
        ----------
        ident: int
            The integer identifier corresponding to an element in the circuit.

        Returns
        -------
        Optional[Element]
        """
        return self._elements.get_element(ident)

    def substitute_element(self, ident: int, element: Element):
        """
        Substitute the element with the given integer identifier in the circuit with another element.

        Parameters
        ----------
        ident: int
            The integer identifier corresponding to an element in the circuit.

        element: Element
            The new element that will substitute the old element.
        """
        assert (
            self._elements.substitute_element(ident, element) is True
        ), f"Failed to find an element with the identifier '{ident}'!"

    def to_sympy(self, substitute: bool = False) -> Expr:
        """
        Get the SymPy expression corresponding to this circuit's impedance.

        Parameters
        ----------
        substitute: bool = False
            Whether or not the variables should be substituted with the current values.

        Returns
        -------
        Expr
        """
        expr: Expr = self._elements.to_sympy(substitute=substitute)
        assert isinstance(expr, Expr)
        return expr

    def to_latex(self) -> str:
        """
        Get the LaTeX math expression corresponding to this circuit's impedance.

        Returns
        -------
        str
        """
        return f"Z = {latex(self.to_sympy(substitute=False))}"

    def to_drawing(
        self,
        node_height: float = 1.5,
        working_label: str = "WE",
        counter_label: str = "CE+RE",
        hide_labels: bool = False,
    ) -> Drawing:
        """
        Get a schemdraw.Drawing object to draw a circuit diagram using the matplotlib backend.

        Parameters
        ----------
        node_height: float = 1.5
            The height of each node.

        working_label: str = "WE"
            The label assigned to the terminal representing the working and working sense electrodes.

        counter_label: str = "CE+RE"
            The label assigned to the terminal representing the counter and reference electrodes.

        hide_labels: bool = False
            Whether or not to hide element and terminal labels.

        Returns
        -------
        Drawing
        """
        lookup: Dict[Type[Element], elm.Element] = {
            Resistor: elm.ResistorIEEE,
            Capacitor: elm.Capacitor,
            ConstantPhaseElement: elm.CPE,
            Inductor: elm.Inductor2,
        }

        def draw_element(elem: Element, drawing: Drawing):
            element: elm.Element = lookup.get(type(elem), elm.ResistorIEC)()
            symbol: str = elem.get_symbol()
            if not hide_labels:
                label: str = elem.get_label()[len(symbol) + 1 :]
                element.label(f"${symbol}_" + r"{\rm " + f"{label}}}$")
            drawing.add(element.right())

        def get_width(
            element_connection: Union[Element, Connection],
        ) -> int:
            widths: List[int] = []
            if isinstance(element_connection, Element):
                widths.append(2)
            elif type(element_connection) is Series:
                for elem_con in element_connection.get_elements(flattened=False):
                    widths.append(
                        get_width(elem_con) + (1 if type(elem_con) is Parallel else 0)
                    )
                widths = [sum(widths)]
            elif type(element_connection) is Parallel:
                for elem_con in element_connection.get_elements(flattened=False):
                    widths.append(get_width(elem_con))
            assert len(widths) > 0
            return max(widths)

        def get_height(element_connection: Union[Element, Connection]) -> int:
            heights: List[int] = []
            if isinstance(element_connection, Element):
                heights.append(node_height)
            elif type(element_connection) is Series:
                for elem_con in element_connection.get_elements(flattened=False):
                    heights.append(get_height(elem_con))
            elif type(element_connection) is Parallel:
                for elem_con in element_connection.get_elements(flattened=False):
                    heights.append(get_height(elem_con))
                heights = [sum(heights)]
            assert len(heights) > 0
            return max(heights)

        def draw_parallel(parallel: Parallel, drawing: Drawing):
            elements_connections: List[Union[Element, Connection]]
            elements_connections = parallel.get_elements(flattened=False)
            heights: List[int] = list(map(get_height, elements_connections))
            i: int
            height: int
            for i, height in enumerate(heights):
                if i < len(elements_connections) - 1:
                    drawing.push()
                    drawing.add(elm.Line(l=height).down())
            total_width: int = get_width(parallel)
            elem_con: Union[Element, Connection]
            for (i, elem_con) in reversed(list(enumerate(elements_connections))):
                width: int = get_width(elem_con)
                padding: int = total_width - width
                if isinstance(elem_con, Connection):
                    if type(elem_con) is Series:
                        draw_series(elem_con, drawing)
                    elif type(elem_con) is Parallel:
                        draw_parallel(elem_con, drawing)
                    else:
                        raise Exception("Unsupported connection: {type(elem_con)=}")
                else:
                    draw_element(elem_con, drawing)
                    if padding > 0:
                        padding += 1
                if padding > 0:
                    drawing.add(elm.Line(l=padding))
                if i > 0:
                    drawing.add(elm.Line(l=heights[i - 1]).up())
                    drawing.pop()

        def draw_series(series: Series, drawing: Drawing):
            i: int
            elem_con: Union[Element, Connection]
            for i, elem_con in enumerate(series.get_elements(flattened=False)):
                if isinstance(elem_con, Connection):
                    if type(elem_con) is Series:
                        draw_series(elem_con, drawing)
                    elif type(elem_con) is Parallel:
                        if i == 0:
                            drawing.add(elm.Line(l=1).right())
                        draw_parallel(elem_con, drawing)
                        if i > 0:
                            drawing.add(elm.Line(l=1))
                    else:
                        raise Exception("Unsupported connection: {type(elem_con)=}")
                else:
                    draw_element(elem_con, drawing)

        drawing: Drawing = Drawing()
        we_dot: elm.Dot = elm.Dot(open=True)
        if not hide_labels:
            we_dot.label(working_label)
        drawing.add(we_dot)
        connection: Connection
        for connection in self.get_connections(flattened=False):
            if type(connection) is Series:
                draw_series(connection, drawing)
            elif type(connection) is Parallel:
                draw_parallel(connection, drawing)
        ce_dot: elm.Dot = elm.Dot(open=True)
        if not hide_labels:
            ce_dot.label(counter_label)
        drawing.add(ce_dot)
        return drawing

    def to_circuitikz(
        self,
        node_width: float = 3.0,
        node_height: float = 1.5,
        working_label: str = "WE",
        counter_label: str = "CE+RE",
        hide_labels: bool = False,
    ) -> str:
        """
        Get the LaTeX source needed to draw a circuit diagram for this circuit using the circuitikz package.

        Parameters
        ----------
        node_width: float = 3.0
            The width of each node.

        node_height: float = 1.5
            The height of each node.

        working_label: str = "WE"
            The label assigned to the terminal representing the working and working sense electrodes.

        counter_label: str = "CE+RE"
            The label assigned to the terminal representing the counter and reference electrodes.

        hide_labels: bool = False
            Whether or not to hide element and terminal labels.

        Returns
        -------
        str
        """
        assert node_width > 0
        assert node_height > 0
        assert type(working_label) is str
        assert type(counter_label) is str
        assert type(hide_labels) is bool
        if hide_labels:
            working_label = ""
            counter_label = ""
        # Phase 1 - figure out the dimensions of the connections and the positions of elements.
        short_counter: int = 0
        dimensions: Dict[
            Union[Series, Parallel, Element, int], Tuple[float, float]
        ] = {}
        positions: Dict[Union[Series, Parallel, Element, int], Tuple[float, float]] = {}
        num_nested_parallels: int = 0

        def short_wire(x: float, y: float) -> Tuple[float, float]:
            nonlocal short_counter
            short_counter += 1
            dimensions[short_counter] = (
                0.25,
                1.0,
            )
            positions[short_counter] = (
                x,
                -y,
            )
            return dimensions[short_counter]

        def phase_1_element(
            element: Element, x: float, y: float
        ) -> Tuple[float, float]:
            dimensions[element] = (
                1.0,
                1.0,
            )
            positions[element] = (
                x,
                -y,
            )
            return dimensions[element]

        def phase_1_series(series: Series, x: float, y: float) -> Tuple[float, float]:
            nonlocal num_nested_parallels
            width: float = 0.0
            height: float = 0.0
            elements: List[Union[Element, Connection]] = series.get_elements(
                flattened=False
            )
            num_elements: int = len(elements)
            i: int
            element_connection: Union[Element, Connection]
            for i, element_connection in enumerate(elements):
                if type(element_connection) is Series:
                    w, h = phase_1_series(element_connection, x + width, y)
                    width += w
                    if h > height:
                        height = h
                elif type(element_connection) is Parallel:
                    if num_nested_parallels > 0 and i == 0:
                        w, h = short_wire(x + width, y)
                        width += w
                        if h > height:
                            height = h
                    w, h = phase_1_parallel(element_connection, x + width, y)
                    width += w
                    if h > height:
                        height = h
                    if num_nested_parallels > 0 and (
                        i == num_elements - 1
                        or (i < num_elements - 1 and type(elements[i + 1]) is Parallel)
                    ):
                        w, h = short_wire(x + width, y)
                        width += w
                        if h > height:
                            height = h
                else:
                    assert isinstance(element_connection, Element)
                    w, h = phase_1_element(element_connection, x + width, y)
                    width += w
            dimensions[series] = (
                max(1, width),
                max(1, height),
            )
            positions[series] = (
                x,
                -y,
            )
            return dimensions[series]

        def phase_1_parallel(
            parallel: Parallel, x: float, y: float
        ) -> Tuple[float, float]:
            nonlocal num_nested_parallels
            num_nested_parallels += 1
            width: float = 0.0
            height: float = 0.0
            for element_connection in parallel.get_elements(flattened=False):
                if type(element_connection) is Series:
                    w, h = phase_1_series(element_connection, x, y + height)
                    if w > width:
                        width = w
                    height += h
                elif type(element_connection) is Parallel:
                    w, h = phase_1_parallel(element_connection, x, y + height)
                    if w > width:
                        width = w
                    height += h
                else:
                    assert isinstance(element_connection, Element)
                    w, h = phase_1_element(element_connection, x, y + height)
                    if w > width:
                        width = w
                    height += h
            dimensions[parallel] = (
                max(1, width),
                max(1, height),
            )
            positions[parallel] = (
                x,
                -y,
            )
            num_nested_parallels -= 1
            return dimensions[parallel]

        assert type(self._elements) is Series
        phase_1_series(self._elements, 0, 0)
        assert set(dimensions.keys()) == set(positions.keys())

        # Phase 2 - generate the LaTeX source for drawing the circuit diagram.
        lines: List[str] = [
            r"\begin{circuitikz}",
            r"\draw (0,0) <label>to[short, o-] (1,0);".replace(
                "<label>",
                f"node[above]{{{working_label}}} " if working_label != "" else "",
            ),
        ]
        line: str
        pos: Tuple[float, float]
        dim: Tuple[float, float]
        symbols: Dict[Type[Element], str] = {
            Resistor: "R",
            Capacitor: "capacitor",
            Inductor: "L",
            ConstantPhaseElement: "cpe",
        }

        def replace_variables(
            line: str,
            start_x: float,
            start_y: float,
            end_x: float,
            end_y: float,
            element: str = "short",
        ) -> str:
            line = line.replace("<start_x>", str(start_x))
            line = line.replace("<start_y>", str(start_y))
            line = line.replace("<end_x>", str(end_x))
            line = line.replace("<end_y>", str(end_y))
            line = line.replace("<element>", element)
            return line

        def phase_2():
            for element_connection in positions:
                x, y = positions[element_connection]
                w, h = dimensions[element_connection]
                if type(element_connection) is Series:
                    continue
                elif type(element_connection) is Parallel:
                    start_x = x * (node_width - 1.0) + 1.0
                    start_y = 1.0
                    end_x = (x + w) * (node_width - 1.0) + 1.0
                    end_y = 1.0
                    for element in dimensions:
                        if not element_connection.contains(element, top_level=True):
                            continue
                        ey = positions[element][1]
                        if start_y > 0.0 or ey > start_y:
                            start_y = ey
                        if end_y > 0.0 or ey < end_y:
                            end_y = ey
                    assert start_y != end_y
                    start_y *= node_height
                    end_y *= node_height
                    line = r"\draw (<start_x>,<start_y>) to[<element>] (<start_x>,<end_y>);"
                    lines.append(
                        replace_variables(line, start_x, start_y, end_x, end_y)
                    )
                    line = r"\draw (<end_x>,<start_y>) to[<element>] (<end_x>,<end_y>);"
                    lines.append(
                        replace_variables(line, start_x, start_y, end_x, end_y)
                    )
                    if w == 1.0:
                        continue
                    for elem_con in filter(
                        lambda _: type(_) is not Parallel, dimensions
                    ):
                        if not element_connection.contains(elem_con, top_level=True):
                            continue
                        if w == dimensions[elem_con][0]:
                            continue
                        if type(elem_con) is Series:
                            ex, ey = positions[elem_con]
                            ew, eh = dimensions[elem_con]
                        else:
                            ex, ey = positions[elem_con]
                            ew, eh = dimensions[elem_con]
                        start_x = (ex + ew) * (node_width - 1.0) + 1.0
                        start_y = ey * node_height
                        # Use the same end_x as the RPar line
                        end_y = start_y
                        line = r"\draw (<start_x>,<start_y>) to[<element>] (<end_x>,<end_y>);"
                        lines.append(
                            replace_variables(line, start_x, start_y, end_x, end_y)
                        )
                elif isinstance(element_connection, Element):
                    start_x = x * (node_width - 1.0) + 1.0
                    start_y = y * node_height
                    end_x = (x + w) * (node_width - 1.0) + 1.0
                    end_y = start_y
                    line = (
                        r"\draw (<start_x>,<start_y>) to[<element>] (<end_x>,<end_y>);"
                    )
                    symbol: str
                    label: str = ""
                    if not hide_labels:
                        symbol, label = element_connection.get_label().split("_")
                        label = f"{symbol}_{{\\rm {label}}}"
                    symbol = symbols.get(type(element_connection), "generic")
                    lines.append(
                        replace_variables(
                            line,
                            max(1.0, start_x),
                            start_y,
                            end_x,
                            end_y,
                            f"{symbol}=${label}$",
                        )
                    )
                elif type(element_connection) is int:
                    start_x = x * (node_width - 1.0) + 1.0
                    start_y = y * node_height
                    end_x = (x + w) * (node_width - 1.0) + 1.0
                    end_y = start_y
                    lines.append(
                        replace_variables(
                            r"\draw (<start_x>,<start_y>) to[short] (<end_x>,<end_y>);",
                            max(1.0, start_x),
                            start_y,
                            end_x,
                            end_y,
                        )
                    )

        phase_2()
        x, y = positions[self._elements]
        w, h = dimensions[self._elements]
        start_x = (x + w) * (node_width - 1) + 1
        end_x = start_x + 1
        line = (
            r"\draw (<start_x>,<start_y>) to[<element>, -o] (<end_x>,<end_y>)<label>;"
        )
        line = line.replace(
            "<label>", f" node[above]{{{counter_label}}}" if counter_label != "" else ""
        )
        lines.append(replace_variables(line, start_x, 0, end_x, 0))
        source: str = "\n\t".join(lines) + "\n\\end{circuitikz}"
        return source



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from typing import (
    Dict,
    Optional,
    Tuple,
)
from numpy import (
    angle,
    array,
    floating,
    issubdtype,
    ndarray,
    sqrt,
    sum as array_sum,
)
from pandas import DataFrame
from scipy.signal import find_peaks


class DRTError(Exception):
    pass


@dataclass(frozen=True)
class DRTResult:
    """
    An object representing the results of calculating the distribution of relaxation times in a  data set.

    Parameters
    ----------
    label: str
        Includes information such as the method used.

    tau: ndarray
        The time constants.

    gamma: ndarray
        The gamma values.
        These values are the real gamma values when the BHT method has been used.

    frequency: ndarray
        The frequencies of the impedance spectrum.

    impedance: ndarray
        The impedance produced by the model.

    real_residual: ndarray
        The residuals of the real parts of the model and the data set.

    imaginary_residual: ndarray
        The residuals of the imaginary parts of the model and the data set.

    mean_gamma: ndarray
        The mean gamma values of the Bayesian credible intervals.

    lower_bound: ndarray
        The lower bound gamma values of the Bayesian credible intervals.

    upper_bound: ndarray
        The upper bound gamma values of the Bayesian credible intervals.

    imaginary_gamma: ndarray
        The imaginary gamma values produced by the BHT method.

    scores: Dict[str, complex]
        The scores calculated by the BHT method.

    chisqr: float
        The chi-squared value of the modeled impedance.

    lambda_value: float
        The lambda value that was ultimately used.
    """

    label: str
    tau: ndarray
    gamma: ndarray
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray
    mean_gamma: ndarray
    lower_bound: ndarray
    upper_bound: ndarray
    imaginary_gamma: ndarray
    scores: Dict[str, complex]
    chisqr: float
    lambda_value: float

    def get_label(self) -> str:
        """
        The label includes information such as the method that was used.

        Returns
        -------
        str
        """
        return self.label

    def get_frequency(self) -> ndarray:
        """
        Get the frequencies (in hertz) of the data set.

        Returns
        -------
        ndarray
        """
        return self.frequency

    def get_impedance(self) -> ndarray:
        """
        Get the complex impedance of the model.

        Returns
        -------
        ndarray
        """
        return self.impedance

    def get_tau(self) -> ndarray:
        """
        Get the time constants.

        Returns
        -------
        ndarray
        """
        return self.tau

    def get_gamma(self, imaginary: bool = False) -> ndarray:
        """
        Get the gamma values.

        Parameters
        ----------
        imaginary: bool = False
            Get the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        ndarray
        """
        if imaginary is True:
            return self.imaginary_gamma
        return self.gamma

    def to_dataframe(
        self,
        threshold: float = 0.0,
        imaginary: bool = False,
        latex_labels: bool = False,
        include_frequency: bool = False,
    ) -> DataFrame:
        """
        Get the peaks as a pandas.DataFrame object that can be used to generate, e.g., a Markdown table.

        Parameters
        ----------
        threshold: float = 0.0
            The threshold for the peaks (0.0 to 1.0 relative to the highest peak).

        imaginary: bool = False
            Use the imaginary gamma (non-empty only when using the BHT method).

        latex_labels: bool = False
            Whether or not to use LaTeX macros in the labels.

        include_frequency: bool = False
            Whether or not to also include a column with the frequencies corresponding to the time constants.

        Returns
        -------
        DataFrame
        """
        tau: ndarray
        gamma: ndarray
        tau, gamma = self.get_peaks(threshold=threshold, imaginary=imaginary)
        f: ndarray = 1 / tau
        dictionary: dict = {}
        dictionary["tau (s)" if not latex_labels else r"$\tau$ (s)"] = tau
        if include_frequency is True:
            dictionary["f (Hz)" if not latex_labels else r"$f$ (Hz)"] = f
        dictionary[
            "gamma (ohms)" if not latex_labels else r"$\gamma\ (\Omega)$"
        ] = gamma
        return DataFrame.from_dict(dictionary)

    def get_peaks(
        self,
        threshold: float = 0.0,
        imaginary: bool = False,
    ) -> Tuple[ndarray, ndarray]:
        """
        Get the time constants (in seconds) and gamma (in ohms) of peaks with magnitudes greater than the threshold.
        The threshold and the magnitudes are all relative to the magnitude of the highest peak.

        Parameters
        ----------
        threshold: float = 0.0
            The threshold for the relative magnitude (0.0 to 1.0).

        imaginary: bool = False
            Use the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        assert (
            issubdtype(type(threshold), floating) and 0.0 <= threshold <= 1.0
        ), threshold
        assert type(imaginary) is bool, imaginary
        gamma: ndarray = self.gamma if not imaginary else self.imaginary_gamma
        assert type(gamma) is ndarray, gamma
        if not gamma.any():
            return (
                array([]),
                array([]),
            )
        indices: ndarray
        indices, _ = find_peaks(gamma)
        if not indices.any():
            return (
                array([]),
                array([]),
            )
        max_g: float = max(gamma)
        if max_g == 0.0:
            return (
                array([]),
                array([]),
            )
        indices = array(
            list(
                filter(
                    lambda _: gamma[_] / max_g > threshold and gamma[_] > 0.0, indices
                )
            )
        )
        if indices.any():
            return (
                self.tau[indices],
                gamma[indices],
            )
        return (
            array([]),
            array([]),
        )

    def get_nyquist_data(self) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Nyquist plot: the real and the negative imaginary parts of the impedances.

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Bode plot: the frequencies, the absolute magnitudes of the impedances, and the negative phase angles/shifts of the impedances in degrees.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_drt_data(self, imaginary: bool = False) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this DRTResult as a DRT plot: the time constants and the corresponding gamma values.

        Parameters
        ----------
        imaginary: bool = False
            Get the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        gamma: ndarray = self.gamma if not imaginary else self.imaginary_gamma
        if not gamma.any():
            return (
                array([]),
                array([]),
            )
        return (
            self.tau,
            gamma,
        )

    def get_drt_credible_intervals(self) -> Tuple[ndarray, ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the Bayesian credible intervals for this DRTResult: the time constants, the mean gamma values, the lower bound gamma values, and the upper bound gamma values.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray, ndarray]
        """
        if not self.mean_gamma.any():
            return (
                array([]),
                array([]),
                array([]),
                array([]),
            )
        return (
            self.tau,
            self.mean_gamma,
            self.lower_bound,
            self.upper_bound,
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the relative residuals for this DRTResult: the frequencies, the relative residuals for the real parts of the impedances in percents, and the relative residuals for the imaginary parts of the impedances in percents.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )

    def get_scores(self) -> Dict[str, complex]:
        """
        Get the scores (BHT method) for the data set.
        The scores are represented as complex values where the real and imaginary parts have magnitudes ranging from 0.0 to 1.0.
        A consistent impedance spectrum should score high.

        Returns
        -------
        Dict[str, complex]
        """
        return self.scores

    def get_score_dataframe(self, latex_labels: bool = False) -> Optional[DataFrame]:
        """
        Get the scores (BHT) method for the data set as a pandas.DataFrame object that can be used to generate, e.g., a Markdown table.

        Parameters
        ----------
        latex_labels: bool = False
            Whether or not to use LaTeX macros in the labels.

        Returns
        -------
        Optional[DataFrame]
        """
        if not self.scores:
            return None
        return DataFrame.from_dict(
            {
                "Score": [
                    "Mean" if not latex_labels else r"$s_\mu$",
                    "Residuals, 1 sigma" if not latex_labels else r"$s_{1\sigma}$",
                    "Residuals, 2 sigma" if not latex_labels else r"$s_{2\sigma}$",
                    "Residuals, 3 sigma" if not latex_labels else r"$s_{3\sigma}$",
                    "Hellinger distance" if not latex_labels else r"$s_{\rm HD}$",
                    "Jensen-Shannon distance" if not latex_labels else r"$s_{\rm JSD}$",
                ],
                ("Real (%)" if not latex_labels else r"Real (\%)"): [
                    self.scores["mean"].real * 100,
                    self.scores["residuals_1sigma"].real * 100,
                    self.scores["residuals_2sigma"].real * 100,
                    self.scores["residuals_3sigma"].real * 100,
                    self.scores["hellinger_distance"].real * 100,
                    self.scores["jensen_shannon_distance"].real * 100,
                ],
                ("Imaginary (%)" if not latex_labels else r"Imaginary (\%)"): [
                    self.scores["mean"].imag * 100,
                    self.scores["residuals_1sigma"].imag * 100,
                    self.scores["residuals_2sigma"].imag * 100,
                    self.scores["residuals_3sigma"].imag * 100,
                    self.scores["hellinger_distance"].imag * 100,
                    self.scores["jensen_shannon_distance"].imag * 100,
                ],
            }
        )


def _calculate_chisqr(Z_exp: ndarray, Z_fit: ndarray) -> float:
    z: complex = array_sum((Z_fit - Z_exp) ** 2 / Z_exp)
    return sqrt(z.real**2 + z.imag**2)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
from collections import OrderedDict
from .base import Element
from numpy import inf


class Resistor(Element):
    """
    Resistor

        Symbol: R

        Z = R

        Variables
        ---------
        R: float = 1E+3 (ohm)
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "R"

    @staticmethod
    def get_description() -> str:
        return "R: Resistor"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "R": 1e3,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "R": False,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "R": 0.0,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "R": inf,
        }

    def impedance(self, f: float) -> complex:
        return complex(self._R, 0)

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "R": self._R,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "R" in parameters:
            self._R = float(parameters["R"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "R"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
from collections import OrderedDict
from math import pi
from numpy import (
    cosh,
    inf,
    sinh,
    tanh,
)
from .base import Element


def coth(x: complex) -> complex:
    return cosh(x) / sinh(x)


class Warburg(Element):
    """
    Warburg (semi-infinite diffusion)

        Symbol: W

        Z = 1/(Y*(2*pi*f*j)^(1/2))

        Variables
        ---------
        Y: float = 1.0 (S*s^(1/2))
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "W"

    @staticmethod
    def get_description() -> str:
        return "W: Warburg, semi-infinite"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "Y": 1.0,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "Y": False,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "Y": 0.0,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "Y": inf,
        }

    def impedance(self, f: float) -> complex:
        return 1 / (self._Y * (2 * pi * f * 1j) ** (1 / 2))

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "Y": self._Y,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "Y" in parameters:
            self._Y = float(parameters["Y"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "1 / (Y * (2*pi*f*I)^(1/2))"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)


class WarburgShort(Element):
    """
    Warburg, finite length or short (finite length diffusion with transmissive boundary)

        Symbol: Ws

        Z = tanh((B*j*2*pi*f)^n)/((Y*j*2*pi*f)^n)

        Variables
        ---------
        Y: float = 1.0 (S)
        B: float = 1.0 (s^n)
        n: float = 0.5
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "Ws"

    @staticmethod
    def get_description() -> str:
        return "Ws: Warburg, finite length or short"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "Y": 1.0,
            "B": 1.0,
            "n": 0.5,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "Y": False,
            "B": False,
            "n": True,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "Y": 0.0,
            "B": 0.0,
            "n": 0.0,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "Y": inf,
            "B": inf,
            "n": 1.0,
        }

    def impedance(self, f: float) -> complex:
        return (
            tanh((self._B * 2 * pi * f * 1j) ** self._n)
            / (self._Y * 2 * pi * f * 1j) ** self._n
        )

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "Y": self._Y,
                "B": self._B,
                "n": self._n,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "Y" in parameters:
            self._Y = float(parameters["Y"])
        if "B" in parameters:
            self._B = float(parameters["B"])
        if "n" in parameters:
            self._n = float(parameters["n"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "tanh((B*I*2*pi*f)^n) / ((Y*I*2*pi*f)^n)"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)


class WarburgOpen(Element):
    """
    Warburg, finite space or open (finite length diffusion with reflective boundary)

        Symbol: Wo

        Z = coth((B*j*2*pi*f)^n)/((Y*j*2*pi*f)^n)

        Variables
        ---------
        Y: float = 1.0 (S)
        B: float = 1.0 (s^n)
        n: float = 0.5
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "Wo"

    @staticmethod
    def get_description() -> str:
        return "Wo: Warburg, finite space or open"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "Y": 1.0,
            "B": 1.0,
            "n": 0.5,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "Y": False,
            "B": False,
            "n": True,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "Y": 0.0,
            "B": 0.0,
            "n": 0.0,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "Y": inf,
            "B": inf,
            "n": 1.0,
        }

    def impedance(self, f: float) -> complex:
        return (
            coth((self._B * 2 * pi * f * 1j) ** self._n)
            / (self._Y * 2 * pi * f * 1j) ** self._n
        )

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "Y": self._Y,
                "B": self._B,
                "n": self._n,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "Y" in parameters:
            self._Y = float(parameters["Y"])
        if "B" in parameters:
            self._B = float(parameters["B"])
        if "n" in parameters:
            self._n = float(parameters["n"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "coth((B*I*2*pi*f)^n) / ((Y*I*2*pi*f)^n)"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import List
from pyimpspec.data import DataSet
from .result import (
    DRTError,
    DRTResult,
)
from .tr_nnls import _calculate_drt_tr_nnls
from .tr_rbf import _calculate_drt_tr_rbf
from .bht import _calculate_drt_bht


_METHODS: List[str] = [
    "bht",
    "tr-nnls",
    "tr-rbf",
]


def calculate_drt(
    data: DataSet,
    method: str = "tr-nnls",
    mode: str = "complex",
    lambda_value: float = -1.0,
    rbf_type: str = "gaussian",
    derivative_order: int = 1,
    rbf_shape: str = "fwhm",
    shape_coeff: float = 0.5,
    inductance: bool = False,
    credible_intervals: bool = False,
    num_samples: int = 2000,
    num_attempts: int = 10,
    maximum_symmetry: float = 0.5,
    num_procs: int = -1,
) -> DRTResult:
    """
    Calculates the distribution of relaxation times (DRT) for a given data set.

    References:

    - Kulikovsky, A., 2020, Phys. Chem. Chem. Phys., 22, 19131-19138 (https://doi.org/10.1039/D0CP02094J)
    - Wan, T. H., Saccoccio, M., Chen, C., and Ciucci, F., 2015, Electrochim. Acta, 184, 483-499 (https://doi.org/10.1016/j.electacta.2015.09.097).
    - Ciucci, F. and Chen, C., 2015, Electrochim. Acta, 167, 439-454 (https://doi.org/10.1016/j.electacta.2015.03.123)
    - Effat, M. B. and Ciucci, F., 2017, Electrochim. Acta, 247, 1117-1129 (https://doi.org/10.1016/j.electacta.2017.07.050)
    - Liu, J., Wan, T. H., and Ciucci, F., 2020, Electrochim. Acta, 357, 136864 (https://doi.org/10.1016/j.electacta.2020.136864)

    Parameters
    ----------
    data: DataSet
        The data set to use in the calculations.

    method: str = "tr-nnls"
        Valid values include:
        - "bht": Bayesian Hilbert Transform
        - "tr-nnls": Tikhonov regularization with non-negative least squares
        - "tr-rbf": Tikhonov regularization with radial basis function discretization

    mode: str = "complex"
        Which parts of the data are to be included in the calculations.
        Used by the "tr-nnls" and "tr-rbf" methods.
        Valid values include:
        - "complex" ("tr-rbf" method only and the default for that method)
        - "real" (default for the "tr-nnls" method)
        - "imaginary"

    lambda_value: float = -1.0
        The Tikhonov regularization parameter.
        Used by the "tr-nnls" and "tr-rbf" methods.
        If the method is "tr-nnls" and this value is equal to or below zero, then an attempt will be made to automatically find a suitable value.

    rbf_type: str = "gaussian"
        The type of function to use for discretization.
        Used by the "bht" and "tr-rbf" methods.
        Valid values include:
        - "gaussian"
        - "c0-matern"
        - "c2-matern"
        - "c4-matern"
        - "c6-matern"
        - "inverse-quadratic"
        - "inverse-quadric"
        - "cauchy"

    derivative_order: int = 1
        The order of the derivative used during discretization.
        Used by the "bht" and "tr-rbf" methods.

    rbf_shape: str = "fwhm"
        The shape control of the radial basis functions.
        Used by the "bht" and "tr-rbf" methods.
        Valid values include:
        - "fwhm": full width half maximum
        - "factor": `shape_coeff` is used directly

    shape_coeff: float = 0.5
        The full width at half maximum (FWHM) coefficient affecting the chosen shape type.
        Used by the "bht" and "tr-rbf" methods.

    inductance: bool = False
        If true, then an inductive element is included in the calculations.
        Used by the "tr-rbf" method.

    credible_intervals: bool = False
        If true, then the credible intervals are also calculated for the DRT results according to Bayesian statistics.
        Used by the "tr-rbf" method.

    num_samples: int = 2000
        The number of samples drawn when calculating the Bayesian credible intervals ("tr-rbf" method) or the Jensen-Shannon distance ("bht" method).
        A greater number provides better accuracy but requires more time.
        Used by the "bht" and "tr-rbf" methods.

    num_attempts: int = 10
        The minimum number of attempts to make when trying to find suitable random initial values.
        A greater number should provide better results at the expense of time.
        Used by the "bht" method.

    maximum_symmetry: float = 0.5
        A maximum limit (between 0.0 and 1.0) for a descriptor of the vertical symmetry of the DRT.
        A high degree of symmetry is common for results where the gamma value oscillates rather than forms distinct peaks.
        A low value for the limit should improve the results but may cause the "bht" method to take longer to finish.
        This limit is only used in the "tr-rbf" method when the regularization parameter (lambda) is not provided.
        Used by the "bht" and "tr-rbf" methods.

    num_procs: int = -1
        The maximum number of processes to use.
        A value below one results in using the total number of CPU cores present.
    """
    assert (
        hasattr(data, "get_frequency")
        and callable(data.get_frequency)
        and hasattr(data, "get_impedance")
        and callable(data.get_impedance)
    ), "Invalid data object!"
    assert (
        type(method) is str and method in _METHODS
    ), f"Unsupported method! Valid values include: {', '.join(_METHODS)}"
    if method == "bht":
        return _calculate_drt_bht(
            data=data,
            rbf_type=rbf_type,
            derivative_order=derivative_order,
            rbf_shape=rbf_shape,
            shape_coeff=shape_coeff,
            num_samples=num_samples,
            num_attempts=num_attempts,
            maximum_symmetry=maximum_symmetry,
            num_procs=num_procs,
        )
    elif method == "tr-nnls":
        if mode == "complex":
            mode = "real"
        return _calculate_drt_tr_nnls(
            data=data,
            mode=mode,
            lambda_value=lambda_value,
            num_procs=num_procs,
        )
    elif method == "tr-rbf":
        return _calculate_drt_tr_rbf(
            data=data,
            mode=mode,
            lambda_value=lambda_value,
            rbf_type=rbf_type,
            derivative_order=derivative_order,
            rbf_shape=rbf_shape,
            shape_coeff=shape_coeff,
            inductance=inductance,
            credible_intervals=credible_intervals,
            num_samples=num_samples,
            maximum_symmetry=maximum_symmetry,
            num_procs=num_procs,
        )
    raise Exception(f"Unsupported method: {method}")



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
from collections import OrderedDict
from math import pi
from numpy import (
    cosh,
    sinh,
)
from .base import Element


def coth(x: complex) -> complex:
    return cosh(x) / sinh(x)


class DeLevieFiniteLength(Element):
    """
    de Levie pore (finite length)

        Symbol: Ls

        Z = (Ri*Rr)^(1/2)*coth(d*(Ri/Rr)^(1/2)*(1+Y*(2*pi*f*j)^n)^(1/2))/(1+Y*(2*pi*f*j)^n)^(1/2)

        Variables
        ---------
        Ri: float = 10.0 (ohm/cm)
        Rr: float = 1.0 (ohm*cm)
        Y: float = 0.01 (F*s^(n-1)/cm)
        n: float = 0.8
        d: float = 0.2 (cm)
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "Ls"

    @staticmethod
    def get_description() -> str:
        return "Ls: de Levie, finite length"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "Ri": 10.0,
            "Rr": 1.0,
            "Y": 0.01,
            "n": 0.8,
            "d": 0.2,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "Ri": False,
            "Rr": False,
            "Y": False,
            "n": False,
            "d": False,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "Ri": 1e-12,
            "Rr": 1e-12,
            "Y": 1e-12,
            "n": 0.5,
            "d": 1e-6,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "Ri": 10.0,
            "Rr": 1.0,
            "Y": 10.0,
            "n": 1.0,
            "d": 5.0,
        }

    def impedance(self, f: float) -> complex:
        return (self._Ri * self._Rr) ** (1 / 2) * (
            coth(
                self._d
                * (self._Ri / self._Rr) ** (1 / 2)
                * (1 + self._Y * (2 * pi * f * 1j) ** self._n) ** (1 / 2)
            )
            / (1 + self._Y * (2 * pi * f * 1j) ** self._n) ** (1 / 2)
        )

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "Ri": self._Ri,
                "Rr": self._Rr,
                "Y": self._Y,
                "n": self._n,
                "d": self._d,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "Ri" in parameters:
            self._Ri = float(parameters["Ri"])
        if "Rr" in parameters:
            self._Rr = float(parameters["Rr"])
        if "Y" in parameters:
            self._Y = float(parameters["Y"])
        if "n" in parameters:
            self._n = float(parameters["n"])
        if "d" in parameters:
            self._d = float(parameters["d"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "sqrt(Ri*Rr)*(coth(d*sqrt(Ri/Rr)*sqrt(1+Y*(2*pi*f*I)^n))/sqrt(1+Y*(2*pi*f*I)^n))"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
from math import pi
from collections import OrderedDict
from .base import Element
from numpy import inf


class HavriliakNegami(Element):
    """
    Havriliak-Negami relaxation

        Symbol: H

        Z = (((1+(j*2*pi*f*t)^a)^b)/(j*2*pi*f*dC))

        Variables
        ---------
        dC: float = 1E-6 (F)
        t: float = 1.0 (s)
        a: float = 0.9
        b: float = 0.9
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "H"

    @staticmethod
    def get_description() -> str:
        return "H: Havriliak-Negami"

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "dC": 1e-6,
            "t": 1.0,
            "a": 0.9,
            "b": 0.9,
        }

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "dC": False,
            "t": False,
            "a": False,
            "b": False,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "dC": 0.0,
            "t": -inf,
            "a": -inf,
            "b": -inf,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "dC": inf,
            "t": inf,
            "a": inf,
            "b": inf,
        }

    def impedance(self, f: float) -> complex:
        return (1 + (2 * pi * f * 1j * self._t) ** self._a) ** self._b / (
            2 * pi * f * 1j * self._dC
        )

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict(
            {
                "dC": self._dC,
                "t": self._t,
                "a": self._a,
                "b": self._b,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        if "dC" in parameters:
            self._dC = float(parameters["dC"])
        if "t" in parameters:
            self._t = float(parameters["t"])
        if "a" in parameters:
            self._a = float(parameters["a"])
        if "b" in parameters:
            self._b = float(parameters["b"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "(((1 + (I*2*pi*f*t)^a)^b) / (I*2*pi*f*dC))"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from collections import OrderedDict
from dataclasses import dataclass
from multiprocessing import (
    Pool,
    cpu_count,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
from traceback import format_exc
import warnings
from numpy import (
    angle,
    array,
    ceil,
    floor,
    inf,
    integer,
    issubdtype,
    log10 as log,
    logspace,
    ndarray,
    ones as ones_array,
    sum as array_sum,
)
from pandas import DataFrame
from lmfit import (
    Parameters,
    minimize,
)
from lmfit.minimizer import MinimizerResult
from pyimpspec.circuit import (
    Circuit,
    parse_cdc,
)
from pyimpspec.circuit.base import (
    Element,
    Connection,
)
from pyimpspec.data import DataSet
import pyimpspec.progress as progress


class FittingError(Exception):
    pass


VERSION: int = 1


@dataclass(frozen=True)
class FittedParameter:
    """
    An object representing a fitted parameter.

    Parameters
    ----------
    value: float
        The fitted value.

    stderr: Optional[float] = None
        The estimated standard error of the fitted value.

    fixed: bool = False
        Whether or not this parameter had a fixed value during the circuit fitting.
    """

    value: float
    stderr: Optional[float] = None
    fixed: bool = False

    @staticmethod
    def _parse_v1(dictionary: dict) -> dict:
        assert type(dictionary) is dict
        return {
            "value": dictionary["value"],
            "stderr": dictionary["stderr"],
            "fixed": dictionary["fixed"],
        }

    @classmethod
    def from_dict(Class, dictionary: dict) -> "FittedParameter":
        assert type(dictionary) is dict, dictionary
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: Class._parse_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        return Class(**parsers[version](dictionary))

    def to_dict(self) -> dict:
        return {
            "version": VERSION,
            "value": self.value,
            "stderr": self.stderr,
            "fixed": self.fixed,
        }

    def get_relative_error(self) -> float:
        return (self.stderr or 0.0) / self.value


def _interpolate(
    experimental: Union[List[float], ndarray], num_per_decade: int
) -> ndarray:
    assert type(experimental) is list or type(experimental) is ndarray, experimental
    assert (
        issubdtype(type(num_per_decade), integer) and num_per_decade > 0
    ), num_per_decade
    min_f: float = min(experimental)
    max_f: float = max(experimental)
    log_min_f: int = int(floor(log(min_f)))
    log_max_f: int = int(ceil(log(max_f)))
    f: float
    freq: List[float] = [
        f
        for f in logspace(
            log_min_f, log_max_f, num=(log_max_f - log_min_f) * num_per_decade + 1
        )
        if f >= min_f and f <= max_f
    ]
    if min_f not in freq:
        freq.append(min_f)
    if max_f not in freq:
        freq.append(max_f)
    return array(list(sorted(freq, reverse=True)))


@dataclass(frozen=True)
class FitResult:
    """
    An object representing the results of fitting a circuit to a data set.

    Parameters
    ----------
    circuit: Circuit
        The fitted circuit.

    parameters: Dict[str, Dict[str, FittedParameter]]
        Fitted parameters and their estimated standard errors (if possible to estimate).

    pseudo_chisqr: float
        The pseudo chi-squared fit value (eq. 14 in Boukamp, 1995).

    minimizer_result: MinimizerResult
        The results of the fit as provided by the lmfit.minimize function.

    frequency: ndarray
        The frequencies used to perform the fit.

    impedance: ndarray
        The impedance produced by the fitted circuit at each of the fitted frequencies.

    real_residual: ndarray
        The residuals for the real parts (eq. 15 in Schönleber et al., 2014).

    imaginary_residual: ndarray
        The residuals for the imaginary parts (eq. 16 in Schönleber et al., 2014).

    method: str
        The iterative method used during the fitting process.

    weight: str
        The weight function used during the fitting process.
    """

    circuit: Circuit
    parameters: Dict[str, Dict[str, FittedParameter]]
    pseudo_chisqr: float
    minimizer_result: MinimizerResult
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray
    method: str
    weight: str

    def __repr__(self) -> str:
        return f"FitResult ({self.circuit.to_string()}, {hex(id(self))})"

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            return _interpolate(self.frequency, num_per_decade)
        return self.frequency

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            return self.circuit.impedances(self.get_frequency(num_per_decade))
        return self.impedance

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this FitResult as a Nyquist plot: the real and the negative imaginary parts of the impedances.

        Parameters
        ----------
        num_per_decade: int = -1
            The number of points per decade.
            A positive value results in data points being calculated using the fitted circuit within the original frequency range.
            Otherwise, only the original frequencies are used.

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                Z.real,
                -Z.imag,
            )
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(
        self, num_per_decade: int = -1
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot this FitResult as a Bode plot: the frequencies, the absolute magnitudes of the impedances, and the negative phase angles/shifts of the impedances in degrees.

        Parameters
        ----------
        num_per_decade: int = -1
            The number of points per decade.
            A positive value results in data points being calculated using the fitted circuit within the original frequency range.
            Otherwise, only the original frequencies are used.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            freq: ndarray = self.get_frequency(num_per_decade)
            Z: ndarray = self.circuit.impedances(freq)
            return (
                freq,
                abs(Z),
                -angle(Z, deg=True),
            )
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the relative residuals for this FitResult: the frequencies, the relative residuals for the real parts of the impedances in percents, and the relative residuals for the imaginary parts of the impedances in percents.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )

    def to_dataframe(self) -> DataFrame:
        element_labels: List[str] = []
        parameter_labels: List[str] = []
        fitted_values: List[float] = []
        stderr_values: List[Optional[float]] = []
        fixed: List[str] = []
        element_label: str
        parameters: Dict[str, FittedParameter]
        elements: List[Element]
        elements = self.circuit.get_elements(flattened=True)  # type: ignore
        element: Union[Element, Connection]
        for element in sorted(
            elements,
            key=lambda _: _.get_identifier(),
        ):
            element_label = element.get_label()
            parameters = self.parameters[element_label]
            parameter_label: str
            parameter: FittedParameter
            for parameter_label, parameter in parameters.items():
                element_labels.append(element_label)
                parameter_labels.append(parameter_label)
                fitted_values.append(parameter.value)
                stderr_values.append(
                    parameter.stderr / parameter.value * 100
                    if parameter.stderr is not None
                    else None
                )
                fixed.append("Yes" if parameter.fixed else "No")
        return DataFrame.from_dict(
            {
                "Element": element_labels,
                "Parameter": parameter_labels,
                "Value": fitted_values,
                "Std. err. (%)": stderr_values,
                "Fixed": fixed,
            }
        )


def _to_lmfit(circuit: Circuit) -> Parameters:
    assert type(circuit) is Circuit, circuit
    result: Parameters = Parameters()
    parameters: "Dict[int, OrderedDict[str, float]]" = circuit.get_parameters()
    ident: int
    params: Dict[str, float]
    for ident, params in parameters.items():
        element: Optional[Element] = circuit.get_element(ident)
        assert element is not None
        element_symbol: str
        symbol: str
        value: float
        for symbol, value in params.items():
            minimum: Any
            maximum: Any
            lower: float = element.get_lower_limit(symbol)
            upper: float = element.get_upper_limit(symbol)
            fixed: bool = element.is_fixed(symbol)
            result.add(f"{symbol}_{ident}", value, min=lower, max=upper, vary=not fixed)
    return result


def _from_lmfit(parameters: Parameters) -> Dict[int, Dict[str, float]]:
    assert type(parameters) is Parameters, parameters
    result: Dict[int, Dict[str, float]] = {}
    key: str
    value: float
    for key, value in parameters.valuesdict().items():
        ident: int
        symbol: str
        symbol, ident = key.split("_")  # type: ignore
        ident = int(ident)
        if ident not in result:
            result[ident] = {}
        result[ident][symbol] = float(value)
    return result


def _residual(
    params: Parameters,
    circuit: Circuit,
    freq: ndarray,
    Z_exp: ndarray,
    weight_func: Callable,
) -> ndarray:
    assert type(params) is Parameters, params
    assert type(circuit) is Circuit, circuit
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    circuit.set_parameters(_from_lmfit(params))
    Z_fit: ndarray = circuit.impedances(freq)
    errors: ndarray = array(
        [(Z_exp.real - Z_fit.real) ** 2, (Z_exp.imag - Z_fit.imag) ** 2]
    )
    return weight_func(Z_exp, Z_fit) * errors


def _unity_weight(Z_exp: ndarray, Z_fit: ndarray) -> ndarray:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    return ones_array(shape=(2, len(Z_exp)))


def _modulus_weight(Z_exp: ndarray, Z_fit: ndarray) -> ndarray:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    return ones_array(shape=(2, len(Z_exp))) / abs(Z_fit)


def _proportional_weight(Z_exp: ndarray, Z_fit: ndarray) -> ndarray:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    weight: ndarray = ones_array(shape=(2, len(Z_exp)))
    weight[0] = weight[0] / Z_fit.real**2
    weight[1] = weight[1] / Z_fit.imag**2
    return weight


def _boukamp_weight(Z_exp: ndarray, Z_fit: ndarray) -> ndarray:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    # See eq. 13 in Boukamp (1995)
    return (Z_exp.real**2 + Z_exp.imag**2) ** -1


_weight_functions: Dict[str, Callable] = {
    "unity": _unity_weight,
    "modulus": _modulus_weight,
    "proportional": _proportional_weight,
    "boukamp": _boukamp_weight,
}


_methods: List[str] = [
    "leastsq",
    "least_squares",
    # "differential_evolution",
    # "brute",
    # "basinhopping",
    # "ampgo",
    "nelder",
    "lbfgsb",
    "powell",
    "cg",
    # "newton",
    # "cobyla",
    "bfgs",
    "tnc",
    # "trust-ncg",
    # "trust-exact",
    # "trust-krylov",
    # "trust-constr",
    # "dogleg",
    "slsqp",
    # "emcee",
    # "shgo",
    # "dual_annealing",
]


def _extract_parameters(
    circuit: Circuit, fit: MinimizerResult
) -> Dict[str, Dict[str, FittedParameter]]:
    assert type(circuit) is Circuit, circuit
    assert type(fit) is MinimizerResult, fit
    parameters: Dict[str, Dict[str, FittedParameter]] = {}
    ident: int
    for ident in reversed(circuit.get_parameters()):
        element: Optional[Element] = circuit.get_element(ident)
        assert element is not None
        label: str = element.get_label()
        assert label not in parameters, label
        parameters[label] = {}
        # Parameters that were not fixed
        name: str
        for name in filter(lambda _: _.endswith(f"_{ident}"), fit.var_names):
            par = fit.params[name]
            stderr: Optional[float] = par.stderr if hasattr(par, "stderr") else None
            parameters[label][name[: name.find("_")]] = FittedParameter(
                par.value, stderr
            )
        # Remaining parameters are fixed
        value: float
        for name, value in element.get_parameters().items():
            if name in parameters[label]:
                continue
            parameters[label][name] = FittedParameter(value, None, True)
    return parameters


def _calculate_pseudo_chisqr(Z_exp: ndarray, Z_fit: ndarray) -> float:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    # See eq. 14 in Boukamp (1995)
    weight: ndarray = _boukamp_weight(Z_exp, Z_fit)
    return float(
        array_sum(
            weight * ((Z_exp.real - Z_fit.real) ** 2 + (Z_exp.imag - Z_fit.imag) ** 2)
        )
    )


def _fit_process(args) -> Tuple[str, Optional[MinimizerResult], float, str, str, str]:
    circuit: Circuit
    freq: ndarray
    Z_exp: ndarray
    method: str
    weight: str
    max_nfev: int
    auto: bool
    circuit, freq, Z_exp, method, weight, max_nfev, auto = args
    assert type(circuit) is Circuit, circuit
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    assert type(method) is str, method
    assert type(weight) is str, weight
    assert issubdtype(type(max_nfev), integer), max_nfev
    assert type(auto) is bool, auto
    weight_func: Callable = _weight_functions[weight]
    with warnings.catch_warnings():
        if auto:
            warnings.filterwarnings("error")
        try:
            fit = minimize(
                _residual,
                _to_lmfit(circuit),
                method,
                args=(circuit, freq, Z_exp, weight_func),
                max_nfev=None if max_nfev < 1 else max_nfev,
            )
        except (Exception, Warning):
            return (
                circuit.to_string(),
                None,
                inf,
                method,
                weight,
                format_exc(),
            )
    if fit.ndata < len(freq) and log(fit.chisqr) < -50:
        return (circuit.to_string(), None, inf, method, weight, "Invalid result!")
    circuit.set_parameters(_from_lmfit(fit.params))
    Z_fit: ndarray = circuit.impedances(freq)
    Xps: float = _calculate_pseudo_chisqr(Z_exp, Z_fit)
    return (
        circuit.to_string(12),
        fit,
        Xps,
        method,
        weight,
        "",
    )


def validate_circuit(circuit: Circuit):
    assert circuit.to_string() not in ["[]", "()"], "The circuit has no elements!"
    element_labels: Set[str] = set()
    for element in circuit.get_elements():
        label: str = element.get_label()
        assert (
            label not in element_labels
        ), f"Two or more elements of the same type have the same label ({label})!"
        element_labels.add(label)


def _calculate_residuals(Z_exp: ndarray, Z_fit: ndarray) -> Tuple[ndarray, ndarray]:
    return (
        (Z_exp.real - Z_fit.real) / abs(Z_exp),
        (Z_exp.imag - Z_fit.imag) / abs(Z_exp),
    )


def fit_circuit(
    circuit: Circuit,
    data: DataSet,
    method: str = "auto",
    weight: str = "auto",
    max_nfev: int = -1,
    num_procs: int = -1,
) -> FitResult:
    """
    Fit a circuit to a data set.

    Parameters
    ----------
    circuit: Circuit
        The circuit to fit to a data set.

    data: DataSet
        The data set that the circuit will be fitted to.

    method: str = "auto"
        The iteration method used during fitting.
        See lmfit's documentation for valid method names.
        Note that not all methods supported by lmfit are possible in the current implementation (e.g. some methods may require a function that calculates a Jacobian).
        The "auto" value results in multiple methods being tested in parallel and the best result being returned based on the chi-squared values.

    weight: str = "auto"
        The weight function to use when calculating residuals.
        Currently supported values: "modulus", "proportional", "unity", "boukamp", and "auto".
        The "auto" value results in multiple weights being tested in parallel and the best result being returned based on the chi-squared values.

    max_nfev: int = -1
        The maximum number of function evaluations when fitting.
        A value less than one equals no limit.

    num_procs: int = -1
        The maximum number of parallel processes to use when method and/or weight is "auto".

    Returns
    -------
    FitResult
    """
    assert type(circuit) is Circuit, (
        type(circuit),
        circuit,
    )
    assert isinstance(data, DataSet), (
        type(data),
        data,
    )
    assert data.get_num_points() > 0
    assert type(method) is str, (
        type(method),
        method,
    )
    assert method in _methods or method == "auto", method
    assert type(weight) is str, (
        type(weight),
        weight,
    )
    assert weight in _weight_functions or weight == "auto", weight
    assert issubdtype(type(max_nfev), integer), (
        type(max_nfev),
        max_nfev,
    )
    assert issubdtype(type(num_procs), integer), (
        type(num_procs),
        num_procs,
    )
    if num_procs < 1:
        num_procs = cpu_count()
    validate_circuit(circuit)
    cdc: str = circuit.to_string(12)
    freq: ndarray = data.get_frequency()
    Z_exp: ndarray = data.get_impedance()
    fits: List[Tuple[str, Optional[MinimizerResult], float, str, str, str]] = []
    methods: List[str] = [method] if method != "auto" else _methods
    weights: List[str] = (
        [weight] if weight != "auto" else list(_weight_functions.keys())
    )
    i: int = 0
    method_weight_combos: List[Tuple[str, str]] = []
    for method in methods:
        for weight in weights:
            method_weight_combos.append((method, weight,))
    args = (
        (
            parse_cdc(cdc),
            freq,
            Z_exp,
            method,
            weight,
            max_nfev,
            True,
        ) for (method, weight) in method_weight_combos
    )
    progress_message = "Performing fit(s)"
    progress.update_every_N_percent(0, message=progress_message)
    try:
        if num_procs > 1:
            with Pool(num_procs) as pool:
                for i, res in enumerate(pool.imap_unordered(_fit_process, args)):
                    progress.update_every_N_percent(
                        i + 1,
                        total=len(method_weight_combos),
                        message=progress_message,
                    )
                    fits.append(res)
        else:
            fits = list(map(_fit_process, args))
        fits.sort(
            key=lambda _: log(_[1].chisqr) + log(_[2]) if _[1] is not None else inf
        )
    except Exception:
        raise FittingError(format_exc())
    if not fits:
        raise FittingError("No valid results generated!")
    fit: Optional[MinimizerResult]
    cdc, fit, Xps, method, weight, error_msg = fits[0]
    if fit is None:
        raise FittingError(error_msg)
    # Return results
    circuit = parse_cdc(cdc)
    Z_fit: ndarray = circuit.impedances(freq)
    return FitResult(
        circuit,
        _extract_parameters(circuit, fit),
        Xps,
        fit,
        freq,
        Z_fit,
        # Residuals calculated according to eqs. 15 and 16
        # in Schönleber et al. (2014)
        *_calculate_residuals(Z_exp, Z_fit),
        method,
        weight,
    )



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from .mpl import (
    plot_bode,
    plot_circuit,
    plot_complex_impedance,
    plot_data,
    plot_drt,
    plot_exploratory_tests,
    plot_fit,
    plot_gamma,
    plot_imaginary_impedance,
    plot_impedance_magnitude,
    plot_impedance_phase,
    plot_mu_xps,
    plot_nyquist,
    plot_real_impedance,
    plot_residual,
)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Optional,
)


_CALLBACK: Optional[Callable] = None


def register(callback: Callable):
    assert callable(callback)
    global _CALLBACK
    _CALLBACK = callback


def update(*args, **kwargs):
    if not callable(_CALLBACK):
        return
    _CALLBACK(*args, **kwargs)


_RECENT_PROGRESS: float = -1.0


def update_every_N_percent(i: int, total: int = 1, N: float = 10.0, *args, **kwargs):
    global _RECENT_PROGRESS
    if i == 0:
        _RECENT_PROGRESS = -1.0
    step: float = N / 100.0
    progress: float = i / total
    if _RECENT_PROGRESS < 0.0 or progress >= _RECENT_PROGRESS + step:
        if _RECENT_PROGRESS < 0.0:
            _RECENT_PROGRESS = 0.0
        else:
            _RECENT_PROGRESS += step
        update(progress=_RECENT_PROGRESS, *args, **kwargs)
    if i >= total:
        _RECENT_PROGRESS = -1.0



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

# This module uses Tikhonov regularization and either radial basis function or piecewise linear discretization
# pyDRTtools commit: 65ea54d9332a0c6594de852f0242a88e20ec4427

from multiprocessing import (
    Pool,
    cpu_count,
)
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)
from numpy import (
    abs as array_abs,
    arccos,
    arctan2,
    argmin,
    array,
    concatenate,
    cos,
    cumsum,
    diff,
    divide,
    exp,
    eye,
    finfo,
    floating,
    inf,
    integer,
    issubdtype,
    log as ln,
    log10 as log,
    logspace,
    mean,
    min as array_min,
    ndarray,
    ones,
    pi,
    quantile,
    real,
    sin,
    sqrt,
    square,
    std,
    sum as array_sum,
    where,
    zeros,
)
from numpy.linalg import cholesky
from numpy.matlib import repmat
from numpy.random import randn
from scipy import integrate
from scipy.linalg import (
    inv as invert,
    solve as solve_linalg,
    toeplitz,
)
from scipy.optimize import (
    fsolve,
)
from cvxopt import (
    matrix,
    solvers as cvxopt_solvers,
)

try:
    from cvxpy import (
        Minimize,
        Problem,
        Variable,
        quad_form,
    )
except ImportError:
    pass
from pyimpspec.data import DataSet
from pyimpspec.analysis.fitting import _calculate_residuals
from .result import (
    DRTError,
    DRTResult,
    _calculate_chisqr,
)
from .tr_nnls import _suggest_lambda
import pyimpspec.progress as progress


RBF_TYPES: List[str] = [
    "c0-matern",
    "c2-matern",
    "c4-matern",
    "c6-matern",
    "cauchy",
    "gaussian",
    "inverse-quadratic",
    "inverse-quadric",
    "piecewise-linear",
]

RBF_SHAPES: List[str] = ["fwhm", "factor"]

MODES: List[str] = ["complex", "real", "imaginary"]


def _generate_truncated_multivariate_gaussians(
    F: ndarray,  # m * d dimensions
    g: ndarray,  # m * 1 dimensions
    M: ndarray,  # d * d dimensions, symmetric and definite positive
    mu_r: ndarray,  # d * 1 dimensions
    initial_X: ndarray,  # d * 1 dimensions
    cov: bool = True,  # True -> M is the covariance and the mean is mu_r, False -> M is a precision matrix (log-density == -1/2 X'*M*X + r'*X)
    L: int = 1,  # Number of samples
) -> ndarray:
    """
    Algorithm described in http://arxiv.org/abs/1208.4118

    F: ndarray
        m * d dimensions

    g: ndarray
        m * 1 dimensions

    M: ndarray
        d * d dimensions, symmetric and definite positive

    mu_r: ndarray
        d * 1 dimensions

    initial_X: ndarray
        d * 1 dimensions

    cov: bool = True
        True -> M is the covariance and the mean is mu_r
        False -> M is a precision matrix (log-density == -1/2 X'*M*X + r'*X)

    L: int = 1
        Number of samples

    Returns an array (d * L dimensions) where each column is a sample
    """
    assert (
        g.shape[0] == F.shape[0]
    ), f"Constraint dimensions do not match: {g.shape[0]} != {F.shape[0]}"
    R: ndarray = cholesky(M)  # .T?
    R = R.T  # change the lower matrix to upper matrix
    mu: ndarray
    if cov is True:  # Using M as a covariance matrix
        mu = mu_r
        g = g + F @ mu
        F = F @ R.T
        initial_X = solve_linalg(R.T, (initial_X - mu))
    else:  # Using M as a precision matrix
        mu = solve_linalg(R, solve_linalg(R.T, mu_r))
        g = g + F @ mu
        F = solve_linalg(R, F)
        initial_X = R @ (initial_X - mu)
    assert (F @ initial_X + g).any() >= 0, "Inconsistent initial condition!"
    # Dimension of mean vector; each sample must be of this dimension
    d: int = initial_X.shape[0]
    near_zero: float = 1e-12
    # Squared Euclidean norm of constraint matrix columns
    F2: ndarray = array_sum(square(F), axis=1)
    Ft: ndarray = F.T
    last_X: ndarray = initial_X
    Xs: ndarray = zeros(
        (
            d,
            L,
        )
    )
    Xs[:, 0] = initial_X
    progress_message: str = "Calculating credible intervals"
    progress.update_every_N_percent(0, message=progress_message)
    # Generate samples
    i: int = 2
    while i <= L:
        if i % 100 == 0:
            progress.update_every_N_percent(i + 1, total=L, message=progress_message)
        stop: bool = False
        j: int = -1
        # Generate initial velocity from normal distribution
        V0: ndarray = randn(d)
        X: ndarray = last_X
        T: float = pi / 2
        tt: float = 0.0
        while True:
            a: ndarray = real(V0)
            b: ndarray = X
            fa: ndarray = F @ a
            fb: ndarray = F @ b
            U: ndarray = sqrt(square(fa) + square(fb))
            # Has to be arctan2 not arctan
            phi: ndarray = arctan2(-fa, fb)
            # Find the locations where the constraints were hit
            pn: ndarray = array(array_abs(divide(g, U)) <= 1)
            if pn.any():
                inds: ndarray = where(pn)[0]
                phn: ndarray = phi[pn]
                t1: ndarray = array_abs(
                    -1.0 * phn + arccos(divide(-1.0 * g[pn], U[pn]))
                )
                # If there was a previous reflection (j > -1) and there is a potential
                # reflection at the sample plane, then make sure that a new reflection
                # at j is not found because of numerical error
                if j > -1:
                    if pn[j] == 1:
                        temp: ndarray = cumsum(pn)
                        indj = temp[j] - 1
                        tt1 = t1[indj]
                        if (
                            array_abs(tt1) < near_zero
                            or array_abs(tt1 - pi) < near_zero
                        ):
                            t1[indj] = inf
                mt: float = array_min(t1)
                m_ind = argmin(t1)
                j = inds[m_ind]
            else:
                mt = T
            # Update travel time
            tt = tt + mt
            if tt >= T:
                mt = mt - (tt - T)
                stop = True
            # Update position and velocity
            X = a * sin(mt) + b * cos(mt)
            V = a * cos(mt) - b * sin(mt)
            if stop:
                break
            # Update new velocity
            qj = F[j, :] @ V / F2[j]
            V0 = V - 2 * qj * Ft[:, j]
        if (F @ X + g).all() > 0:
            Xs[:, i - 1] = X
            last_X = X
            i = i + 1
    if cov:
        Xs = R.T @ Xs + repmat(mu.reshape(mu.shape[0], 1), 1, L)
    else:
        Xs = solve_linalg(R, Xs) + repmat(mu.reshape(mu.shape[0], 1), 1, L)
    return Xs


def _calculate_credible_intervals(
    num_RL: int,
    num_samples: int,
    mu: ndarray,
    Sigma_inv: ndarray,
    x: ndarray,
    tau_fine: ndarray,
    tau: ndarray,
    epsilon: float,
    rbf_type: str,
) -> Tuple[ndarray, ndarray, ndarray]:
    # Calculation of credible interval according to Bayesian statistics
    mu = mu[num_RL:]
    Sigma_inv = Sigma_inv[num_RL:, num_RL:]
    # Cholesky transform instead of direct inverse
    L_Sigma_inv: ndarray = cholesky(Sigma_inv)
    L_Sigma_agm: ndarray = invert(L_Sigma_inv)
    Sigma: ndarray = L_Sigma_agm.T @ L_Sigma_agm
    # Using generate_tmg from HMC_exact.py to sample the truncated Gaussian distribution
    Xs: ndarray = _generate_truncated_multivariate_gaussians(
        eye(x.shape[0]),
        finfo(float).eps * ones(mu.shape[0]),
        Sigma,
        mu,
        x,
        True,
        num_samples,
    )
    lower_bound: ndarray
    upper_bound: ndarray
    # map array to gamma
    _, lower_bound = _x_to_gamma(
        quantile(Xs[:, 501:], 0.005, axis=1),
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    )
    _, upper_bound = _x_to_gamma(
        quantile(Xs[:, 501:], 0.995, axis=1),
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    )
    _, mean_gamma = _x_to_gamma(
        mean(Xs[:, 501:], axis=1),
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    )
    return (
        mean_gamma,
        lower_bound,
        upper_bound,
    )


def _rbf_epsilon_functions(func: Callable) -> Callable:
    switch: Dict[str, Callable] = {
        "gaussian": lambda x: exp(-((x) ** 2)) - 0.5,
        "c0-matern": lambda x: exp(-abs(x)) - 0.5,
        "c2-matern": lambda x: exp(-abs(x)) * (1 + abs(x)) - 0.5,
        "c4-matern": lambda x: 1 / 3 * exp(-abs(x)) * (3 + 3 * abs(x) + abs(x) ** 2)
        - 0.5,
        "c6-matern": lambda x: 1
        / 15
        * exp(-abs(x))
        * (15 + 15 * abs(x) + 6 * abs(x) ** 2 + abs(x) ** 3)
        - 0.5,
        "inverse-quadratic": lambda x: 1 / (1 + (x) ** 2) - 0.5,
        "inverse-quadric": lambda x: 1 / sqrt(1 + (x) ** 2) - 0.5,
        "cauchy": lambda x: 1 / (1 + abs(x)) - 0.5,
        "piecewise-linear": lambda x: 0.0,
    }
    assert set(RBF_TYPES) == set(switch.keys())

    def wrapper(*args, **kwargs):
        kwargs["rbf_functions"] = switch
        return func(*args, **kwargs)

    return wrapper


@_rbf_epsilon_functions
def _compute_epsilon(
    f: ndarray,
    rbf_shape: str,
    shape_coeff: float,
    rbf_type: str,
    rbf_functions: Dict[str, Callable],
) -> float:
    if rbf_type == "piecewise-linear":
        return 0.0
    elif rbf_shape == "fwhm":
        FWHM_coeff: ndarray = 2 * fsolve(rbf_functions[rbf_type], 1)
        delta: float = mean(diff(ln(1 / f.reshape(f.shape[0]))))
        return (shape_coeff * FWHM_coeff / delta)[0]
    # "factor"
    return shape_coeff


def _rbf_A_matrix_functions(func: Callable) -> Callable:
    switch: Dict[str, Callable] = {
        "c0-matern": lambda x, epsilon: exp(-abs(epsilon * x)),
        "c2-matern": lambda x, epsilon: exp(-abs(epsilon * x)) * (1 + abs(epsilon * x)),
        "c4-matern": lambda x, epsilon: 1
        / 3
        * exp(-abs(epsilon * x))
        * (3 + 3 * abs(epsilon * x) + abs(epsilon * x) ** 2),
        "c6-matern": lambda x, epsilon: 1
        / 15
        * exp(-abs(epsilon * x))
        * (
            15
            + 15 * abs(epsilon * x)
            + 6 * abs(epsilon * x) ** 2
            + abs(epsilon * x) ** 3
        ),
        "cauchy": lambda x, epsilon: 1 / (1 + abs(epsilon * x)),
        "gaussian": lambda x, epsilon: exp(-((epsilon * x) ** 2)),
        "inverse-quadratic": lambda x, epsilon: 1 / (1 + (epsilon * x) ** 2),
        "inverse-quadric": lambda x, epsilon: 1 / sqrt(1 + (epsilon * x) ** 2),
        "piecewise-linear": lambda x, epsilon: 0.0,
    }
    assert set(RBF_TYPES) == set(switch.keys())

    def wrapper(*args, **kwargs):
        kwargs["rbf_functions"] = switch
        return func(*args, **kwargs)

    return wrapper


@_rbf_A_matrix_functions
def _A_matrix_element(
    f: float,
    tau: float,
    epsilon: float,
    real: bool,
    rbf_type: str,
    rbf_functions: Dict[str, Callable],
) -> float:
    alpha: float = 2 * pi * f * tau
    rbf_func: Callable = rbf_functions[rbf_type]
    integrand: Callable
    if real is True:
        integrand = (
            lambda x: 1.0 / (1.0 + (alpha**2) * exp(2.0 * x)) * rbf_func(x, epsilon)
        )
    else:
        integrand = (
            lambda x: alpha
            / (1.0 / exp(x) + (alpha**2) * exp(x))
            * rbf_func(x, epsilon)
        )
    return integrate.quad(integrand, -50, 50, epsabs=1e-9, epsrel=1e-9)[0]


def _assemble_A_matrix(args) -> ndarray:
    f: ndarray
    tau: ndarray
    epsilon: float
    real: bool
    rbf_type: str
    (
        f,
        tau,
        epsilon,
        real,
        rbf_type,
    ) = args
    w: ndarray = 2 * pi * f
    num_freqs: int = f.shape[0]
    num_taus: int = tau.shape[0]
    A: ndarray
    i: int
    j: int
    # Check if the frequencies are ln-spaced
    if (
        num_freqs == num_taus
        and (std(diff(ln(1 / f))) / mean(diff(ln(1 / f)))) < 0.01
        and rbf_type != "piecewise-linear"
    ):
        # Use the Toeplitz trick
        C: ndarray = zeros(num_freqs)
        for i in range(0, num_freqs):
            C[i] = _A_matrix_element(f[i], tau[0], epsilon, real, rbf_type)
        R: ndarray = zeros(num_taus)
        for j in range(0, num_taus):
            R[j] = _A_matrix_element(f[0], tau[j], epsilon, real, rbf_type)
        A = toeplitz(C, R)
    else:
        # Use brute force
        A = zeros(
            (
                num_freqs,
                num_taus,
            )
        )
        for i in range(0, num_freqs):
            for j in range(0, num_taus):
                if rbf_type == "piecewise-linear":
                    if real is True:
                        A[i, j] = (
                            0.5
                            / (1 + (w[i] * tau[j]) ** 2)
                            * ln(
                                (tau[j] if j == num_taus - 1 else tau[j + 1])
                                / (tau[j] if j == 0 else tau[j - 1])
                            )
                        )
                    else:
                        A[i, j] = (
                            # -0.5
                            0.5
                            * (w[i] * tau[j])
                            / (1 + (w[i] * tau[j]) ** 2)
                            * ln(
                                (tau[j] if j == num_taus - 1 else tau[j + 1])
                                / (tau[j] if j == 0 else tau[j - 1])
                            )
                        )
                else:
                    A[i, j] = _A_matrix_element(f[i], tau[j], epsilon, real, rbf_type)
    return (1 if real is True else -1) * A


def _inner_product_rbf(
    f_i: float,
    f_j: float,
    epsilon: float,
    derivative_order: int,
    rbf_type: str,
) -> float:
    a: float = epsilon * ln(f_i / f_j)
    if rbf_type == "c0-matern":
        if derivative_order == 1:
            return epsilon * (1 - abs(a)) * exp(-abs(a))
        elif derivative_order == 2:
            return epsilon**3 * (1 + abs(a)) * exp(-abs(a))
    elif rbf_type == "c2-matern":
        if derivative_order == 1:
            return epsilon / 6 * (3 + 3 * abs(a) - abs(a) ** 3) * exp(-abs(a))
        elif derivative_order == 2:
            return (
                epsilon**3
                / 6
                * (3 + 3 * abs(a) - 6 * abs(a) ** 2 + abs(a) ** 3)
                * exp(-abs(a))
            )
    elif rbf_type == "c4-matern":
        if derivative_order == 1:
            return (
                epsilon
                / 30
                * (
                    105
                    + 105 * abs(a)
                    + 30 * abs(a) ** 2
                    - 5 * abs(a) ** 3
                    - 5 * abs(a) ** 4
                    - abs(a) ** 5
                )
                * exp(-abs(a))
            )
        elif derivative_order == 2:
            return (
                epsilon**3
                / 30
                * (45 + 45 * abs(a) - 15 * abs(a) ** 3 - 5 * abs(a) ** 4 + abs(a) ** 5)
                * exp(-abs(a))
            )
    elif rbf_type == "c6-matern":
        if derivative_order == 1:
            return (
                epsilon
                / 140
                * (
                    10395
                    + 10395 * abs(a)
                    + 3780 * abs(a) ** 2
                    + 315 * abs(a) ** 3
                    - 210 * abs(a) ** 4
                    - 84 * abs(a) ** 5
                    - 14 * abs(a) ** 6
                    - abs(a) ** 7
                )
                * exp(-abs(a))
            )
        elif derivative_order == 2:
            return (
                epsilon**3
                / 140
                * (
                    2835
                    + 2835 * abs(a)
                    + 630 * abs(a) ** 2
                    - 315 * abs(a) ** 3
                    - 210 * abs(a) ** 4
                    - 42 * abs(a) ** 5
                    + abs(a) ** 7
                )
                * exp(-abs(a))
            )
    elif rbf_type == "cauchy":
        if a == 0:
            if derivative_order == 1:
                return 2 / 3 * epsilon
            elif derivative_order == 2:
                return 8 / 5 * epsilon**3
        else:
            numerator: float
            denominator: float
            if derivative_order == 1:
                numerator = abs(a) * (2 + abs(a)) * (
                    4 + 3 * abs(a) * (2 + abs(a))
                ) - 2 * (1 + abs(a)) ** 2 * (4 + abs(a) * (2 + abs(a))) * ln(1 + abs(a))
                denominator = abs(a) ** 3 * (1 + abs(a)) * (2 + abs(a)) ** 3
                return 4 * epsilon * numerator / denominator
            elif derivative_order == 2:
                numerator = abs(a) * (2 + abs(a)) * (
                    -96
                    + abs(a)
                    * (2 + abs(a))
                    * (-30 + abs(a) * (2 + abs(a)))
                    * (4 + abs(a) * (2 + abs(a)))
                ) + 12 * (1 + abs(a)) ^ 2 * (
                    16 + abs(a) * (2 + abs(a)) * (12 + abs(a) * (2 + abs(a)))
                ) * ln(
                    1 + abs(a)
                )
                denominator = abs(a) ^ 5 * (1 + abs(a)) * (2 + abs(a)) ** 5
                return 8 * epsilon ^ 3 * numerator / denominator
    elif rbf_type == "gaussian":
        if derivative_order == 1:
            return -epsilon * (-1 + a**2) * exp(-(a**2 / 2)) * sqrt(pi / 2)
        elif derivative_order == 2:
            return (
                epsilon**3
                * (3 - 6 * a**2 + a**4)
                * exp(-(a**2 / 2))
                * sqrt(pi / 2)
            )
    elif rbf_type == "inverse-quadratic":
        if derivative_order == 1:
            return 4 * epsilon * (4 - 3 * a**2) * pi / ((4 + a**2) ** 3)
        elif derivative_order == 2:
            return (
                48
                * (16 + 5 * a**2 * (-8 + a**2))
                * pi
                * epsilon**3
                / ((4 + a**2) ** 5)
            )
    elif rbf_type == "inverse-quadric":
        y_i: float = -ln(f_i)
        y_j: float = -ln(f_j)
        rbf_i: Callable = lambda y: 1 / sqrt(1 + (epsilon * (y - y_i)) ** 2)
        rbf_j: Callable = lambda y: 1 / sqrt(1 + (epsilon * (y - y_j)) ** 2)
        delta: float
        sqr_drbf_dy: Callable
        if derivative_order == 1:
            delta = 1e-8
            sqr_drbf_dy = (
                lambda y: 1
                / (2 * delta)
                * (rbf_i(y + delta) - rbf_i(y - delta))
                * 1
                / (2 * delta)
                * (rbf_j(y + delta) - rbf_j(y - delta))
            )
        elif derivative_order == 2:
            delta = 1e-4
            sqr_drbf_dy = (
                lambda y: 1
                / (delta ^ 2)
                * (rbf_i(y + delta) - 2 * rbf_i(y) + rbf_i(y - delta))
                * 1
                / (delta ^ 2)
                * (rbf_j(y + delta) - 2 * rbf_j(y) + rbf_j(y - delta))
            )
        return integrate.quad(sqr_drbf_dy, -50, 50, epsabs=1e-9, epsrel=1e-9)[0]
    assert rbf_type not in RBF_TYPES, f"Unsupported RBF type: {rbf_type}"
    return -1.0  # Just to satisfy mypy


def _assemble_M_matrix(
    tau: ndarray,
    epsilon: float,
    derivative_order: int,
    rbf_type: str,
) -> ndarray:
    f: ndarray = 1 / tau
    num_freqs: int = f.shape[0]
    num_taus: int = tau.shape[0]
    M: ndarray
    i: int
    j: int
    # Check if the collocation points are sufficiently ln-spaced
    if (
        std(diff(ln(tau))) / mean(diff(ln(tau)))
    ) < 0.01 and rbf_type != "piecewise-linear":
        # Apply the Toeplitz trick to compute the M matrix
        C: ndarray = zeros(num_taus)
        for i in range(0, num_taus):
            C[i] = _inner_product_rbf(
                f[0],
                f[i],
                epsilon,
                derivative_order,
                rbf_type,
            )  # TODO: Maybe use tau instead of freq (pyDRTtools comment)
        R: ndarray = zeros(num_taus)
        for j in range(0, num_taus):
            R[j] = _inner_product_rbf(
                f[j],
                f[0],
                epsilon,
                derivative_order,
                rbf_type,
            )
        M = toeplitz(C, R)
    elif rbf_type == "piecewise-linear":
        if derivative_order == 1:
            M = zeros(
                (
                    num_freqs - 1,
                    num_freqs,
                )
            )
            for i in range(0, num_freqs - 1):
                delta_loc: float = ln((1 / f[i + 1]) / (1 / f[i]))
                M[i, i] = -1 / delta_loc
                M[i, i + 1] = 1 / delta_loc
        elif derivative_order == 2:
            M = zeros(
                (
                    num_taus - 2,
                    num_taus,
                )
            )
            for i in range(0, num_taus - 2):
                delta_loc = ln(tau[i + 1] / tau[i])

                if i == 0 or i == num_taus - 3:
                    M[i, i] = 2.0 / (delta_loc**2)
                    M[i, i + 1] = -4.0 / (delta_loc**2)
                    M[i, i + 2] = 2.0 / (delta_loc**2)
                else:
                    M[i, i] = 1.0 / (delta_loc**2)
                    M[i, i + 1] = -2.0 / (delta_loc**2)
                    M[i, i + 2] = 1.0 / (delta_loc**2)
        M = M.T @ M
    else:
        # Brute force
        M = zeros(
            (
                num_taus,
                num_taus,
            )
        )
        for i in range(0, num_taus):
            for j in range(0, num_taus):
                M[i, j] = _inner_product_rbf(
                    f[i],
                    f[j],
                    epsilon,
                    derivative_order,
                    rbf_type,
                )  # TODO: Maybe use tau instead of freq? See previous pyDRTtools comment.
    return M


def _quad_format(
    A: ndarray,
    b: ndarray,
    M: ndarray,
    lambda_value: float,
) -> Tuple[ndarray, ndarray]:
    H: ndarray = 2 * (A.T @ A + lambda_value * M)
    H = (H.T + H) / 2
    c: ndarray = -2 * b.T @ A
    return (
        H,
        c,
    )


def _quad_format_combined(
    A_re: ndarray,
    A_im: ndarray,
    b_re: ndarray,
    b_im: ndarray,
    M: ndarray,
    lambda_value: float,
) -> Tuple[ndarray, ndarray]:
    H: ndarray = 2 * ((A_re.T @ A_re + A_im.T @ A_im) + lambda_value * M)
    H = (H.T + H) / 2
    c: ndarray = -2 * (b_im.T @ A_im + b_re.T @ A_re)
    return (
        H,
        c,
    )


def _solve_qp(
    H: ndarray,
    c: ndarray,
    G: Optional[ndarray] = None,
    h: Optional[ndarray] = None,
    A: Optional[ndarray] = None,
    b: Optional[ndarray] = None,
) -> ndarray:
    try:
        # cvxpy (optional dependency)
        N_out: int = c.shape[0]
        x: Variable = Variable(shape=N_out, value=ones(N_out))
        l: ndarray = zeros(N_out)
        prob: Problem = Problem(Minimize((1 / 2) * quad_form(x, H) + c @ x), [x >= l])
        prob.solve(
            # verbose=True,
            eps_abs=1e-10,
            eps_rel=1e-10,
            sigma=1.00e-08,
            max_iter=200000,
            eps_prim_inf=1e-5,
            eps_dual_inf=1e-5,
        )
        return x.value
    except Exception:
        # cvxopt
        args: List[matrix] = [matrix(H), matrix(c)]
        if G is not None:
            assert h is not None
            args.extend([matrix(G), matrix(h)])
        if A is not None:
            assert b is not None
            args.extend([matrix(A), matrix(b)])
        cvxopt_solvers.options["abstol"] = 1e-15
        cvxopt_solvers.options["reltol"] = 1e-15
        solution: dict = cvxopt_solvers.qp(*args)
        if "optimal" not in solution["status"]:
            raise DRTError("Failed to find optimal solution!")
        return array(solution["x"]).reshape((H.shape[1],))


def _rbf_gamma_functions(func: Callable) -> Callable:
    switch: Dict[str, Callable] = {
        "c0-matern": lambda x, epsilon: exp(-abs(epsilon * x)),
        "c2-matern": lambda x, epsilon: exp(-abs(epsilon * x)) * (1 + abs(epsilon * x)),
        "c4-matern": lambda x, epsilon: 1
        / 3
        * exp(-abs(epsilon * x))
        * (3 + 3 * abs(epsilon * x) + abs(epsilon * x) ** 2),
        "c6-matern": lambda x, epsilon: 1
        / 15
        * exp(-abs(epsilon * x))
        * (
            15
            + 15 * abs(epsilon * x)
            + 6 * abs(epsilon * x) ** 2
            + abs(epsilon * x) ** 3
        ),
        "cauchy": lambda x, epsilon: 1 / (1 + abs(epsilon * x)),
        "gaussian": lambda x, epsilon: exp(-((epsilon * x) ** 2)),
        "inverse-quadratic": lambda x, epsilon: 1 / (1 + (epsilon * x) ** 2),
        "inverse-quadric": lambda x, epsilon: 1 / sqrt(1 + (epsilon * x) ** 2),
        "piecewise-linear": lambda x, epsilon: 0.0,
    }
    assert set(RBF_TYPES) == set(switch.keys())

    def wrapper(*args, **kwargs):
        kwargs["rbf_functions"] = switch
        return func(*args, **kwargs)

    return wrapper


@_rbf_gamma_functions
def _x_to_gamma(
    x: ndarray,
    tau_fine: ndarray,
    tau: ndarray,
    epsilon: float,
    rbf_type: str,
    rbf_functions: Dict[str, Callable],
) -> Tuple[ndarray, ndarray]:
    # TODO: double check this to see if the function is correct (pyDRTtools comment)
    if rbf_type == "piecewise-linear":
        return (
            tau,
            x,
        )
    num_taus: int = tau.shape[0]
    num_fine_taus: int = tau_fine.shape[0]
    B: ndarray = zeros(
        (
            num_fine_taus,
            num_taus,
        )
    )
    rbf: Callable = rbf_functions[rbf_type]
    i: int
    j: int
    for i in range(0, num_fine_taus):
        for j in range(0, num_taus):
            delta_ln_tau = ln(tau_fine[i]) - ln(tau[j])
            B[i, j] = rbf(delta_ln_tau, epsilon)
    return (
        tau_fine,
        B @ x,
    )


def _prepare_complex_matrices(
    A_re: ndarray,
    A_im: ndarray,
    b_re: ndarray,
    b_im: ndarray,
    M: ndarray,
    lambda_value: float,
    f: ndarray,
    num_freqs: int,
    num_taus: int,
    inductance: bool,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, int]:
    num_RL: int = 2 if inductance is True else 1
    tmp: ndarray  # Used for temporary binding of matrices
    tmp = A_re
    A_re = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_re[:, num_RL:] = tmp
    A_re[:, 1 if inductance is True else 0] = 1

    tmp = A_im
    A_im = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_im[:, num_RL:] = tmp
    if inductance is True:
        A_im[:, 0] = 2 * pi * f

    tmp = M
    M = zeros(
        (
            num_taus + num_RL,
            num_taus + num_RL,
        )
    )
    M[num_RL:, num_RL:] = tmp

    H, c = _quad_format_combined(
        A_re,
        A_im,
        b_re,
        b_im,
        M,
        lambda_value,
    )
    return (
        A_re,
        A_im,
        M,
        H,
        c,
        num_RL,
    )


def _prepare_real_matrices(
    A_re: ndarray,
    A_im: ndarray,
    b_re: ndarray,
    b_im: ndarray,
    M: ndarray,
    lambda_value: float,
    num_freqs: int,
    num_taus: int,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, int]:
    num_RL: int = 1
    tmp: ndarray  # Used for temporary binding of matrices
    tmp = A_re
    A_re = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_re[:, num_RL:] = tmp
    A_re[:, 0] = 1

    tmp = A_im
    A_im = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_im[:, num_RL:] = tmp

    tmp = M
    M = zeros(
        (
            num_taus + num_RL,
            num_taus + num_RL,
        )
    )
    M[num_RL:, num_RL:] = tmp

    H, c = _quad_format(
        A_re,
        b_re,
        M,
        lambda_value,
    )
    return (
        A_re,
        A_im,
        M,
        H,
        c,
        num_RL,
    )


def _prepare_imaginary_matrices(
    A_re: ndarray,
    A_im: ndarray,
    b_re: ndarray,
    b_im: ndarray,
    M: ndarray,
    lambda_value: float,
    f: ndarray,
    num_freqs: int,
    num_taus: int,
    inductance: bool,
) -> Tuple[ndarray, ndarray, ndarray, ndarray, ndarray, int]:
    num_RL: int = 1 if inductance is True else 0
    tmp: ndarray  # Used for temporary binding of matrices
    tmp = A_re
    A_re = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_re[:, num_RL:] = tmp

    tmp = A_im
    A_im = zeros(
        (
            num_freqs,
            num_taus + num_RL,
        )
    )
    A_im[:, num_RL:] = tmp
    if inductance is True:
        A_im[:, 0] = 2 * pi * f
    tmp = M
    M = zeros(
        (
            num_taus + num_RL,
            num_taus + num_RL,
        )
    )
    M[num_RL:, num_RL:] = tmp

    H, c = _quad_format(
        A_im,
        b_im,
        M,
        lambda_value,
    )
    return (
        A_re,
        A_im,
        M,
        H,
        c,
        num_RL,
    )


def _lambda_process(
    args,
) -> Optional[
    Tuple[
        float,
        float,
        ndarray,
        ndarray,
        ndarray,
        ndarray,
        ndarray,
        ndarray,
        ndarray,
        int,
    ]
]:
    A_re: ndarray
    A_im: ndarray
    Z: ndarray
    M: ndarray
    lambda_value: float
    f: ndarray
    tau: ndarray
    tau_fine: ndarray
    epsilon: float
    mode: str
    rbf_type: str
    inductance: bool
    maximum_symmetry: float
    (
        A_re,
        A_im,
        Z,
        M,
        lambda_value,
        f,
        tau,
        tau_fine,
        epsilon,
        mode,
        rbf_type,
        inductance,
        maximum_symmetry,
    ) = args
    num_freqs: int = f.size
    num_taus: int = tau.size
    num_RL: int  # The number of R and/or L elements in series
    H: ndarray
    c: ndarray
    if mode == "complex":
        A_re, A_im, M, H, c, num_RL = _prepare_complex_matrices(
            A_re,
            A_im,
            Z.real,
            Z.imag,
            M,
            lambda_value,
            f,
            num_freqs,
            num_taus,
            inductance,
        )
    elif mode == "real":
        A_re, A_im, M, H, c, num_RL = _prepare_real_matrices(
            A_re,
            A_im,
            Z.real,
            Z.imag,
            M,
            lambda_value,
            num_freqs,
            num_taus,
        )
    elif mode == "imaginary":
        A_re, A_im, M, H, c, num_RL = _prepare_imaginary_matrices(
            A_re,
            A_im,
            Z.real,
            Z.imag,
            M,
            lambda_value,
            f,
            num_freqs,
            num_taus,
            inductance,
        )
    try:
        x: ndarray = _solve_qp(H, c)
    except Exception:
        return None
    gamma: ndarray
    _, gamma = _x_to_gamma(x[num_RL:], tau_fine, tau, epsilon, rbf_type)
    min_gamma: float = abs(min(gamma))
    max_gamma: float = abs(max(gamma))
    score: float = 1.0 - ((max_gamma - min_gamma) / max(min_gamma, max_gamma))
    if score > maximum_symmetry:
        return None
    Z_fit: ndarray = array(
        list(map(lambda _: complex(_[0], _[1]), zip(A_re @ x, A_im @ x)))
    )
    return (
        lambda_value,
        sqrt(array_sum(gamma**2)),
        x,
        Z_fit,
        A_re,
        A_im,
        M,
        H,
        c,
        num_RL,
    )


def _calculate_drt_tr_rbf(
    data: DataSet,
    mode: str = "complex",
    lambda_value: float = 1e-3,
    rbf_type: str = "gaussian",
    derivative_order: int = 1,
    rbf_shape: str = "fwhm",
    shape_coeff: float = 0.5,
    inductance: bool = False,
    credible_intervals: bool = False,
    num_samples: int = 2000,
    maximum_symmetry: float = 0.3,
    num_procs: int = -1,
) -> DRTResult:
    assert type(mode) is str and mode in MODES, f"Valid mode values: {', '.join(MODES)}"
    assert issubdtype(type(lambda_value), floating), lambda_value
    assert (
        type(rbf_type) is str and rbf_type in RBF_TYPES
    ), f"Valid rbf_type values: {', '.join(RBF_TYPES)}"
    assert (
        issubdtype(type(derivative_order), integer) and 1 <= derivative_order <= 2
    ), "Valid derivative_order values: 1, 2"
    assert (
        type(rbf_shape) is str and rbf_shape in RBF_SHAPES
    ), f"Valid rbf_shape values: {', '.join(RBF_SHAPES)}"
    assert (
        issubdtype(
            type(shape_coeff),
            floating,
        )
        and shape_coeff > 0.0
    ), shape_coeff
    assert type(inductance) is bool, inductance
    assert type(credible_intervals) is bool, credible_intervals
    if credible_intervals is True:
        assert num_samples >= 1000, f"{num_samples} is not enough samples!"
    assert issubdtype(type(num_samples), integer) and num_samples > 0, num_samples
    assert (
        issubdtype(
            type(maximum_symmetry),
            floating,
        )
        and 0.0 <= maximum_symmetry <= 1.0
    ), maximum_symmetry
    assert issubdtype(type(num_procs), integer), num_procs
    if num_procs < 1:
        num_procs = cpu_count()
    min_log_lambda: float = -15.0
    max_log_lambda: float = 0.0
    lambda_values: ndarray = (
        logspace(
            min_log_lambda,
            max_log_lambda,
            num=round((max_log_lambda - min_log_lambda)) * 1 + 1,
        )
        if lambda_value <= 0.0
        else array([lambda_value])
    )
    # TODO: Figure out if f and Z need to be altered depending on the value
    # of the 'inductance' argument!
    f: ndarray = data.get_frequency()
    Z: ndarray = data.get_impedance()
    tau: ndarray = 1 / f
    tau_fine: ndarray = logspace(
        log(tau.min()) - 0.5, log(tau.max()) + 0.5, 10 * f.shape[0]
    )
    num_freqs: int = f.size
    epsilon: float = _compute_epsilon(f, rbf_shape, shape_coeff, rbf_type)
    progress_message: str = "Preparing matrices"
    progress.update_every_N_percent(0, total=3, message=progress_message)
    A_re: ndarray
    A_im: ndarray
    with Pool(2 if num_procs > 1 else 1) as pool:
        for i, res in enumerate(
            pool.imap(
                _assemble_A_matrix,
                [
                    (
                        f,
                        tau,
                        epsilon,
                        True,
                        rbf_type,
                    ),
                    (
                        f,
                        tau,
                        epsilon,
                        False,
                        rbf_type,
                    ),
                ],
            )
        ):
            progress.update_every_N_percent(i + 1, total=3, message=progress_message)
            if i == 0:
                A_re = res
            else:
                A_im = res
    M: ndarray = _assemble_M_matrix(tau, epsilon, derivative_order, rbf_type)
    progress.update_every_N_percent(1, message=progress_message)
    with Pool(num_procs) as pool:
        args = (
            (
                A_re,
                A_im,
                Z,
                M,
                lambda_value,
                f,
                tau,
                tau_fine,
                epsilon,
                mode,
                rbf_type,
                inductance,
                maximum_symmetry if lambda_values.size > 1 else 1.0,
            )
            for lambda_value in lambda_values
        )
        results: List[
            Tuple[
                float,
                float,
                ndarray,
                ndarray,
                ndarray,
                ndarray,
                ndarray,
                ndarray,
                ndarray,
                int,
            ]
        ] = []
        progress_message = "Calculating DRT"
        progress.update_every_N_percent(0, message=progress_message)
        i: int
        for i, res in enumerate(pool.imap_unordered(_lambda_process, args)):
            progress.update_every_N_percent(
                i + 1,
                total=len(lambda_values),
                message=progress_message,
            )
            if res is not None:
                results.append(res)
    if len(results) == 0:
        raise DRTError("Failed to perform calculations! Try tweaking the settings.")
    if len(results) > 1:
        results.sort(key=lambda _: _[0])
        lambda_value = _suggest_lambda(
            array(list(map(lambda _: _[0], results))),
            array(list(map(lambda _: _[1], results))),
        )
        results = list(filter(lambda _: _[0] == lambda_value, results))
    lambda_value, _, x, Z_fit, A_re, A_im, M, H, c, num_RL = results[0]
    sigma_re_im: float
    if mode == "complex":
        sigma_re_im = std(concatenate([Z_fit.real - Z.real, Z_fit.imag - Z.imag]))
    elif mode == "real":
        sigma_re_im = std(Z_fit.real - Z.real)
    elif mode == "imaginary":
        sigma_re_im = std(Z_fit.imag - Z.imag)
    inv_V: ndarray = 1 / sigma_re_im**2 * eye(num_freqs)
    Sigma_inv: ndarray
    mu_numerator: ndarray
    if mode == "complex":
        Sigma_inv = (
            (A_re.T @ inv_V @ A_re)
            + (A_im.T @ inv_V @ A_im)
            + (lambda_value / sigma_re_im**2) * M
        )
        mu_numerator = A_re.T @ inv_V @ Z.real + A_im.T @ inv_V @ Z.imag
    elif mode == "real":
        Sigma_inv = (A_re.T @ inv_V @ A_re) + (lambda_value / sigma_re_im**2) * M
        mu_numerator = A_re.T @ inv_V @ Z.real
    elif mode == "imaginary":
        Sigma_inv = (A_im.T @ inv_V @ A_im) + (lambda_value / sigma_re_im**2) * M
        mu_numerator = A_im.T @ inv_V @ Z.imag
    Sigma_inv = (Sigma_inv + Sigma_inv.T) / 2
    L_Sigma_inv: ndarray = cholesky(Sigma_inv)
    mu: ndarray = solve_linalg(L_Sigma_inv.T, solve_linalg(L_Sigma_inv, mu_numerator))
    # TODO: Why were L and R defined only to not be used?
    L: float
    R: float
    if num_RL == 0:
        L, R = 0.0, 0.0
    elif num_RL == 1:
        if mode == "imaginary":
            L, R = x[0], 0.0
        else:
            L, R = 0.0, x[0]
    elif num_RL == 2:
        L, R = x[0:2]
    x = x[num_RL:]
    return DRTResult(
        "TR-RBF",
        *_x_to_gamma(x, tau_fine, tau, epsilon, rbf_type),
        f,
        Z_fit,
        *_calculate_residuals(Z, Z_fit),
        # "tr-rbf" method with credible_intervals
        *(
            _calculate_credible_intervals(
                num_RL,
                num_samples,
                mu,
                Sigma_inv,
                x,
                tau_fine,
                tau,
                epsilon,
                rbf_type,
            )
            if credible_intervals is True
            else (
                array([]),  # Mean
                array([]),  # Lower bound
                array([]),  # Upper bound
            )
        ),
        # "bht" method
        array([]),  # Imaginary gamma
        {},
        # Stats
        _calculate_chisqr(Z, Z_fit),
        lambda_value,
    )



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from numpy import inf
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)
from .base import (
    Connection,
    Element,
)
from .parallel import Parallel
from .series import Series
from .kk import KramersKronigRC
from .resistor import Resistor
from .capacitor import Capacitor
from .inductor import Inductor
from .constant_phase_element import ConstantPhaseElement
from .gerischer import Gerischer
from .havriliak_negami import HavriliakNegami
from .warburg import (
    Warburg,
    WarburgOpen,
    WarburgShort,
)
from .de_levie import DeLevieFiniteLength
from .circuit import Circuit
from .tokenizer import (
    Tokenizer,
    Token,
    Identifier,
    Number,
    FixedNumber,
    LBracket,
    RBracket,
    LCurly,
    RCurly,
    LParen,
    RParen,
    Comma,
    Percent,
    Equals,
    ForwardSlash,
    Colon,
    Label,
)


class ParsingError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class InsufficientTokens(ParsingError):
    def __init__(self):
        super().__init__("Ran out of tokens!")


class UnexpectedToken(ParsingError):
    def __init__(self, token: Token, Class: Optional[Type[Token]] = None):
        super().__init__(f"Unexpected token '{token.value}'!")
        self.token: Token = token
        self.expected: Optional[Type[Token]] = Class


class ExpectedParameterIdentifier(ParsingError):
    def __init__(self, Class: Type[Element]):
        super().__init__(f"Expected a parameter identifier for '{Class.get_symbol()}'!")
        self.element: Type[Element] = Class


class ExpectedNumericValue(ParsingError):
    def __init__(self, token: Optional[Token]):
        super().__init__("Expected numeric value!")
        self.token: Optional[Token] = token


class ConnectionWithoutElements(ParsingError):
    def __init__(self, Class: Type[Connection]):
        super().__init__(
            ("Series" if Class is Series else "Parallel")
            + " connection closed without any elements or connections!"
        )
        self.connection: Type[Connection] = Class


class InsufficientElementsInParallelConnection(ParsingError):
    def __init__(self):
        super().__init__(
            "Parallel connections must contain at least two elements and/or connections!"
        )


class InvalidElementSymbol(ParsingError):
    def __init__(self, identifier: Identifier):
        super().__init__(f"Invalid element symbol '{identifier.value}'!")
        self.identifier: Identifier = identifier


class DuplicateParameterDefinition(ParsingError):
    def __init__(self, key: str, remaining_keys: List[str], Class: Type[Element]):
        super().__init__(
            f"Duplicate parameter definition ('{key}') for '{Class.get_symbol()}'!"
        )
        self.key: str = key
        self.remaining_keys: List[str] = remaining_keys
        self.element: Type[Element] = Class


class InvalidParameterDefinition(ParsingError):
    def __init__(self, key: str, valid_keys: List[str], Class: Type[Element]):
        super().__init__(
            f"Invalid parameter definition ('{key}') for '{Class.get_symbol()}'!"
        )
        self.key: str = key
        self.valid_keys: List[str] = valid_keys
        self.element: Type[Element] = Class


class TooManyParameterDefinitions(ParsingError):
    def __init__(self, Class: Type[Element]):
        super().__init__(f"Too many parameter definitions for '{Class.get_symbol()}'!")
        self.element: Type[Element] = Class


class InvalidParameterLowerLimit(ParsingError):
    def __init__(self, identifier: Identifier, value: float, limit: float):
        super().__init__(f"Invalid lower limit '{limit:.3E}' for value '{value:.3E}'!")
        self.identifier: Identifier = identifier
        self.value: float = value
        self.limit: float = limit


class InvalidParameterUpperLimit(ParsingError):
    def __init__(self, identifier: Identifier, value: float, limit: float):
        super().__init__(f"Invalid upper limit '{limit:.3E}' for value '{value:.3E}'!")
        self.identifier: Identifier = identifier
        self.value: float = value
        self.limit: float = limit


Stackable = Union[Token, Element, Connection]


class Parser:
    def __init__(self):
        self._tokens: List[Token] = []
        self._stack: List[Stackable] = []
        Class: Type[Element]
        self._valid_elements: Dict[str, Type[Element]] = {
            Class.get_symbol(): Class
            for Class in [
                KramersKronigRC,
                Resistor,
                Capacitor,
                Inductor,
                ConstantPhaseElement,
                Gerischer,
                HavriliakNegami,
                Warburg,
                WarburgOpen,
                WarburgShort,
                DeLevieFiniteLength,
            ]
        }
        aliases: Dict[str, Type[Element]] = {
            "T": WarburgOpen,
            "O": WarburgShort,
            "D": DeLevieFiniteLength,
        }
        for symbol in aliases:
            assert (
                symbol not in self._valid_elements
            ), f"Symbol '{symbol}' already in use by {self._valid_elements[symbol]}"
        self._valid_elements.update(aliases)

    def process(self, string: str) -> Circuit:
        string = string.strip()
        if string == "" or string == "[]":
            self.push_stack(Series([]))
        else:
            tokenizer: Tokenizer = Tokenizer()
            tokens: List[Token] = tokenizer.process(string)
            if tokens:
                self._tokens = tokens
                while self._tokens:
                    self.main_loop()
        con: Union[Series, Parallel]
        if self.get_stack_length() > 1:
            elements: List[Union[Element, Connection]] = []
            while not self.is_stack_empty():
                elem: Union[Element, Connection] = self.pop_stack()  # type: ignore
                assert isinstance(elem, Element) or isinstance(elem, Connection)
                elements.append(elem)
            con = Series(elements)
        else:
            con = self.pop_stack()  # type: ignore
            if type(con) is not Series:
                con = Series([con])
        assert self.is_stack_empty() is True, self._stack
        return Circuit(con)

    def pop_token(self) -> Token:
        if len(self._tokens) == 0:
            raise InsufficientTokens()
        return self._tokens.pop(0)

    def pop_stack(self) -> Stackable:
        assert len(self._stack) > 0, "Ran out of items on the stack!"
        return self._stack.pop(0)

    def push_stack(self, item: Stackable):
        self._stack.insert(0, item)

    def is_stack_empty(self) -> bool:
        return len(self._stack) == 0

    def get_stack_length(self) -> int:
        return len(self._stack)

    def peek(self, n: int = 1) -> Optional[Token]:
        assert n >= 0, n
        if len(self._tokens) < n + 1:
            return None
        return self._tokens[n]

    def accept(self, Class: Type[Token]) -> bool:
        if not self._tokens:
            return False
        return type(self._tokens[0]) is Class

    def expect(self, Class: Type[Token]):
        if len(self._tokens) == 0:
            raise InsufficientTokens()
        token: Token = self._tokens[0]
        if type(token) is not Class:
            raise UnexpectedToken(token, Class)

    def expect_number(self):
        if len(self._tokens) == 0:
            raise ExpectedNumericValue(None)
        token: Token = self._tokens[0]
        if not (type(token) is Number or type(token) is FixedNumber):
            raise ExpectedNumericValue(token)

    def main_loop(self):
        if self.accept(LBracket):
            self.connection(LBracket, RBracket, Series)
        elif self.accept(LParen):
            self.connection(LParen, RParen, Parallel)
        elif self.accept(Identifier):
            self.element()
        elif self._tokens:
            raise UnexpectedToken(self.pop_token())
        else:
            raise InsufficientTokens()

    def connection(
        self, Opening: Type[Token], Closing: Type[Token], Class: Type[Connection]
    ):
        self.push_stack(self.pop_token())
        if self.accept(Closing):
            raise ConnectionWithoutElements(Class)
        while not self.accept(Closing):
            self.main_loop()
        self.expect(Closing)
        self.pop_token()
        items: List[Stackable] = []
        while self._stack:
            item: Stackable = self.pop_stack()
            if type(item) is Opening:
                break
            if type(item) is Class:
                items.extend(item._elements)
            else:
                items.append(item)
        assert len(items) > 0
        if Class is Parallel and len(items) < 2:
            raise InsufficientElementsInParallelConnection()
        assert all(
            map(lambda _: isinstance(_, Element) or isinstance(_, Connection), items)
        )
        if Class is Series and len(items) == 1:
            self.push_stack(items[0])
        else:
            self.push_stack(Class(items))  # type: ignore

    def element(self):
        identifier: Identifier = self.pop_token()
        if identifier.value not in self._valid_elements:
            raise InvalidElementSymbol(identifier)
        Class: Type[Element] = self._valid_elements[identifier.value]
        label: str
        parameters: Dict[str, float]
        lower_limits: Dict[str, float]
        upper_limits: Dict[str, float]
        fixed_parameters: Dict[str, bool]
        (
            label,
            parameters,
            lower_limits,
            upper_limits,
            fixed_parameters,
        ) = self.parameters(Class)
        element = Class(**parameters)
        element.set_label(label)
        key: str
        value: Union[float, bool]
        for key, value in lower_limits.items():
            element.set_lower_limit(key, value)
        for key, value in upper_limits.items():
            element.set_upper_limit(key, value)
        for key, value in fixed_parameters.items():
            element.set_fixed(key, value)
        self.push_stack(element)

    def parameters(
        self, Class: Type[Element]
    ) -> Tuple[
        str, Dict[str, float], Dict[str, float], Dict[str, float], Dict[str, bool]
    ]:
        label: str = ""
        parameters: Dict[str, float] = {}
        lower_limits: Dict[str, float] = {}
        upper_limits: Dict[str, float] = {}
        fixed_parameters: Dict[str, bool] = {}
        defaults: Dict[str, float] = Class.get_defaults()
        if not self.accept(LCurly):
            return (
                label,
                defaults,
                lower_limits,
                upper_limits,
                fixed_parameters,
            )
        self.pop_token()
        if not self.accept(Colon):
            keys: List[str] = list(defaults.keys())
            key: str
            value: float
            while keys:
                key, value, lower, upper, fixed = self.param(Class)
                if key in parameters:
                    raise DuplicateParameterDefinition(key, keys, Class)
                if key not in keys:
                    raise InvalidParameterDefinition(key, keys, Class)
                keys.remove(key)
                del defaults[key]
                parameters[key] = value
                lower_limits[key] = lower
                upper_limits[key] = upper
                fixed_parameters[key] = fixed
                if self.accept(Comma):
                    if len(keys) == 0:
                        raise TooManyParameterDefinitions(Class)
                    self.pop_token()
                    continue
                break
        if self.accept(Colon):
            self.pop_token()
            self.expect(Label)
            label = self.pop_token().value
        self.expect(RCurly)
        self.pop_token()
        for key, value in defaults.items():
            if key not in parameters:
                parameters[key] = value
        return (
            label,
            parameters,
            lower_limits,
            upper_limits,
            fixed_parameters,
        )

    def param(self, Class: Type[Element]) -> Tuple[str, float, float, float, bool]:
        if not self.accept(Identifier):
            raise ExpectedParameterIdentifier(Class)
        key: Identifier
        key = self.pop_token()  # type: ignore
        self.expect(Equals)
        self.pop_token()
        self.expect_number()
        value: Union[Number, FixedNumber]
        value = self.pop_token()  # type: ignore
        fixed: bool = False
        lower: float = -inf
        upper: float = inf
        if type(value) is FixedNumber:
            fixed = True
        elif self.accept(ForwardSlash):
            self.pop_token()
            if self.accept(ForwardSlash):
                self.pop_token()
                upper = self.param_limit(value.value)
            else:
                lower = self.param_limit(value.value)
                if self.accept(ForwardSlash):
                    self.pop_token()
                    upper = self.param_limit(value.value)
        if lower > value.value:
            raise InvalidParameterLowerLimit(key, value.value, lower)
        if upper < value.value:
            raise InvalidParameterUpperLimit(key, value.value, upper)
        return (
            key.value,
            value.value,
            lower,
            upper,
            fixed,
        )

    def param_limit(self, value: float) -> float:
        self.expect(Number)
        limit: Number
        limit = self.pop_token()  # type: ignore
        if self.accept(Percent):
            self.pop_token()
            return value * limit.value / 100
        return limit.value



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
from math import pi
from collections import OrderedDict
from .base import Element
from numpy import inf


class KramersKronigRC(Element):
    """
    Parallel RC element used in linear Kramers-Kronig tests

        Symbol: K

        Z = R / (1 + j*2*pi*f*t)

        Variables
        ---------
        R: float = 1 (ohm)
        t: float = 1 (s)
    """

    def __init__(self, **kwargs):
        keys: List[str] = list(self.get_defaults().keys())
        super().__init__(keys=keys)
        self.reset_parameters(keys)
        self.set_parameters(kwargs)

    @staticmethod
    def get_symbol() -> str:
        return "K"

    @staticmethod
    def get_description() -> str:
        return "K: Linear Kramers-Kronig test element"

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        return {
            "R": False,
            "t": True,
        }

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        return {
            "R": 1,
            "t": 1,
        }

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        return {
            "R": -inf,
            "t": 0,
        }

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        return {
            "R": inf,
            "t": inf,
        }

    def impedance(self, f: float) -> complex:
        return self._R / (1 + 2 * pi * f * self._t * 1j)

    def get_parameters(self) -> "OrderedDict[str, float]":
        return OrderedDict({"R": self._R, "t": self._t})

    def set_parameters(self, parameters: Dict[str, float]):
        if "R" in parameters:
            self._R = float(parameters["R"])
        if "t" in parameters:
            self._t = float(parameters["t"])

    def _str_expr(self, substitute: bool = False) -> str:
        string: str = "R / (1 + I*2*pi*f*t)"
        return self._subs_str_expr(string, self.get_parameters(), not substitute)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
    Type,
    Union,
)
from pandas import DataFrame
from numpy import (
    logspace,
    ndarray,
)
from pyimpspec.data import DataSet
from .base import (
    Connection,
    Element,
)
from .parser import (
    Parser,
    ParsingError,
)
from .circuit import Circuit
from .parallel import Parallel
from .series import Series
from .resistor import Resistor
from .capacitor import Capacitor
from .inductor import Inductor
from .constant_phase_element import ConstantPhaseElement
from .gerischer import Gerischer
from .havriliak_negami import HavriliakNegami
from .warburg import (
    Warburg,
    WarburgOpen,
    WarburgShort,
)
from .de_levie import DeLevieFiniteLength


def get_elements() -> Dict[str, Type[Element]]:
    """
    Returns a mapping of element symbols to the element class.

    Returns
    -------
    Dict[str, Type[Element]]
    """
    return {
        _.get_symbol(): _
        for _ in sorted(
            [
                Resistor,
                Capacitor,
                Inductor,
                ConstantPhaseElement,
                Warburg,
                WarburgShort,
                WarburgOpen,
                DeLevieFiniteLength,
                Gerischer,
                HavriliakNegami,
            ],
            key=lambda _: _.get_symbol(),
        )
    }


def parse_cdc(cdc: str) -> Circuit:
    """
    Generate a Circuit instance from a string that contains a circuit description code (CDC).

    Parameters
    ----------
    cdc: str
        A circuit description code (CDC) corresponding to an equivalent circuit.

    Returns
    -------
    Circuit
    """
    assert type(cdc) is str
    return Parser().process(cdc)


class CircuitBuilder:
    """
    A class for building circuits using context managers

    Parameters
    ----------
    parallel: bool = False
        Whether or not this context/connection is a parallel connection.
    """

    def __init__(self, parallel: bool = False):
        self._is_parallel: bool = parallel
        self._elements: List[Union["CircuitBuilder", Element]] = []

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        if self._is_parallel:
            assert (
                len(self._elements) >= 2
            ), "Parallel connections must contain at least two items (elements and/or other connections)."
        else:
            assert (
                len(self._elements) >= 1
            ), "Series connections must contain at least one item (an element or another connection)."

    def __str__(self) -> str:
        return self.to_string()

    def series(self) -> "CircuitBuilder":
        """
        Create a series connection.

        Returns
        -------
        CircuitBuilder
        """
        series: "CircuitBuilder" = CircuitBuilder(parallel=False)
        self._elements.append(series)
        return series

    def parallel(self) -> "CircuitBuilder":
        """
        Create a parallel connection.

        Returns
        -------
        CircuitBuilder
        """
        parallel: "CircuitBuilder" = CircuitBuilder(parallel=True)
        self._elements.append(parallel)
        return parallel

    def __iadd__(self, element: Element) -> "CircuitBuilder":
        self.add(element)
        return self

    def add(self, element: Element):
        """
        Add an element to the current context (i.e., connection).

        Parameters
        ----------
        element: Element
            The element to add to the current series or parallel connection.
        """
        assert isinstance(element, Element), element
        self._elements.append(element)

    def _to_string(self, decimals: int = 12) -> str:
        cdc: str = "(" if self._is_parallel else "["
        element: Union["CircuitBuilder", Element]
        for element in self._elements:
            if isinstance(element, Element):
                cdc += element.to_string(decimals=decimals)
            else:
                cdc += element._to_string(decimals=decimals)
        cdc += ")" if self._is_parallel else "]"
        return cdc

    def to_string(self, decimals: int = -1) -> str:
        """
        Generate a circuit description code.

        Parameters
        ----------
        decimals: int = -1
            The number of decimals to include for the current element parameter values and limits.
            -1 means that the CDC is generated using the basic syntax, which omits element labels, parameter values, and parameter limits.

        Returns
        -------
        str
        """
        return self.to_circuit().to_string(decimals=decimals)

    def to_circuit(self) -> Circuit:
        """
        Generate a circuit.

        Returns
        -------
        Circuit
        """
        return parse_cdc(self._to_string())


def simulate_spectrum(
    circuit: Circuit,
    frequencies: Union[List[float], ndarray] = [],
    label: str = "",
) -> DataSet:
    """
    Simulate the impedance spectrum generated by a circuit in a certain frequency range.

    Parameters
    ----------
    circuit: Circuit
        The circuit to use when calculating impedances at various frequencies.

    frequencies: Union[List[float], ndarray] = []
        A list of floats representing frequencies in Hz.
        If no frequencies are provided, then a frequency range of 10 mHz to 100 kHz with 10 points per decade will be used.

    label: str = ""
        The label for the DataSet that is returned.

    Returns
    -------
    DataSet
    """
    assert type(circuit) is Circuit
    assert type(frequencies) is list or type(frequencies) is ndarray
    assert type(label) is str
    if len(frequencies) == 0:
        frequencies = logspace(5, -2, 71)
    columns: dict = {
        "frequency": frequencies,
        "real": [],
        "imaginary": [],
        "label": label,
    }
    f: float
    for f in frequencies:
        z: complex = circuit.impedance(f)
        columns["real"].append(z.real)
        columns["imaginary"].append(z.imag)
    return DataSet.from_dict(columns)



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import exists
from typing import (
    IO,
    List,
)
from pyimpspec.data.data_set import (
    DataFrame,
    DataSet,
    dataframe_to_dataset,
)


def parse_mpt(path: str) -> DataSet:
    """
    Parse a BioLogic EC-Lab .mpt file containing an impedance spectrum.

    Parameters
    ----------
    path: str
        The path to the file to process.

    Returns
    -------
    DataSet
    """
    assert type(path) is str and exists(path)
    fp: IO
    with open(path, "r", encoding="latin1") as fp:
        lines: List[str] = list(
            filter(lambda _: _ != "", map(str.lower, map(str.strip, fp.readlines())))
        )
    assert lines.pop(0) == "ec-lab ascii file"
    while lines:
        line: str = lines.pop(0)
        if line.startswith("freq/hz"):
            break
    freq: List[float] = []
    real: List[float] = []
    imag: List[float] = []
    while lines:
        columns: List[str] = lines.pop(0).split("\t")
        assert len(columns) > 3, (
            len(columns),
            columns,
        )
        freq.append(float(columns[0]))
        real.append(float(columns[1]))
        imag.append(-float(columns[2]))
    assert len(freq) == len(real) == len(imag) > 0, len(freq)
    return dataframe_to_dataset(
        DataFrame.from_dict(
            {
                "frequency": freq,
                "real": real,
                "imaginary": imag,
            }
        ),
        path,
        "",
    )
